/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.cache;

// import java.awt.BufferCapabilities.FlipContents;
import java.io.File;
import java.util.Date;

import casa.dodwan.config.DodwanEnvironment;
import casa.dodwan.docware.Descriptor;
import casa.dodwan.docware.DescriptorSelector;
import casa.dodwan.docware.DescriptorZipBufferizer;
import casa.dodwan.message.Message;
import casa.dodwan.util.Bufferizer;
import casa.dodwan.util.Console;
// import casa.dodwan.util.Dispatcher;
import casa.dodwan.util.Service;
import casa.dodwan.util.Sink;

/**
 * The Class CacheService.
 */
public class CacheService extends Service {
    
    public Cache cache;
    
    private Console console_ = null;

    /**
     * Instantiates a new cache service.
     * @param tService the transmission service
     * @param path the path to cache
     * @throws Exception
     */

    public CacheService(File path)
	throws Exception {
	
	// ------------------------------------------------------------
	// CACHE MANAGEMENT
	cache = new Cache(path);
	
	// ------------------------------------------------------------
	// LOAD CACHE FROM FILESYSTEM
// 	verbosity.enable();
	if (path != null) {
	    casa.dodwan.util.Logger.log("cache", "loading from filesystem");
	    load(path, false);
	}
// 	verbosity.disable();

	// ------------------------------------------------------------
	// LOG CURRENT FILE STATUS
	int capacity   = DodwanEnvironment.getInstance().cache_capacity;
	long used      = cache.getWeight();
	int nbElts     = cache.getNbElements();

	if (capacity > 0) {
	    long free = capacity - used;
	    casa.dodwan.util.Logger.log("cache", "status: nb=" + nbElts
					+ ",capacity=" + capacity
					+ ",used=" + used
					+ "(" + (100 * used / capacity) 
					+ "%),free=" + free
					+ "(" + (100 * free / capacity)
					+ "%)");
	}
	else {
	    casa.dodwan.util.Logger.log("cache", "status: nb=" + nbElts
					+ ",used=" + used);
	}
    }

    // ------------------------------------------------------------
    /*
     * (non-Javadoc)
     * @see casa.dodxwan.util.Service#console()
     */
    @Override
	public Console console() {
	if (console_ == null)
	    console_ = new CacheServiceConsole(this);
	return console_;
    }
    
    // ------------------------------------------------------------
    /**
     * Delete keys from cache.
     * @param keys the keys to delete
     * @throws Exception
     */
    public void delete(String[] keys) throws Exception {
	
	for (int i = 1; i < keys.length; i++) {
	    cache.remove(keys[i]);
	}
    }
    
    // ------------------------------------------------------------
    /*
     * (non-Javadoc)
     * @see casa.dodwan.util.Suspendable#disable()
     */
    @Override
	public void disable() {
	
	if (verbosity.isEnabled())
	    System.out.println("CacheService.disable()");
	
	if (isEnabled()) {
	    super.disable();
	}
	else if (verbosity.isEnabled())
	    System.out.println("  Failure: this service is not enabled");
    }
    
    // ------------------------------------------------------------
    /*
     * (non-Javadoc)
     * @see casa.dodwan.util.Suspendable#enable()
     */
    @Override
	public void enable() {
	
	if (verbosity.isEnabled())
	    System.out.println("CacheService.enable()");
	
	if (!isEnabled()) {
	    super.enable();
	}
	else if (verbosity.isEnabled())
	    System.out.println("  Failure: this service is already enabled");
    }
    
    // ------------------------------------------------------------
    /**
     * Load the cache.
     * @param path the path to cache
     * @param flush
     * @throws Exception
     */
    public void load(File path, boolean flush) throws Exception {
	
	if (verbosity.isEnabled())
	    System.out.println("CacheService.load(" 
			       + path.getAbsoluteFile() 
			       + ")");
	
	File dir = new File(path, "descriptors");
	File[] files;
	if (dir.isDirectory())
	    files = dir.listFiles();
	else {
	    files = new File[1];
	    files[0] = dir;
	}
	
	Bufferizer<Descriptor> bufferizer = new DescriptorZipBufferizer();
	long now = System.currentTimeMillis();
	for (int i = 0; i < files.length; i++) {
	    File file = files[i];

	    if (verbosity.isEnabled())
		System.out.println("Loading " + file.getAbsoluteFile());
	    try {
		Descriptor desc = bufferizer.fromFile(file);
		String key = desc.getDocumentId();

		// Check if message is obsolete.  If not put it
		// in cache. Otherwise remove it from
		// filesystem
		boolean obsolete = false;
		Date deadline = 
		    desc.getDeadline();
		if (deadline != null) {
		    long lifetime = deadline.getTime() - now;
		    obsolete = (lifetime <= 0);
		}
		
		// Check if message is cancelled
		String status = desc.getAttribute("_cancelled");
		boolean cancelled = ((status != null) && (status.equals("true")));

		if (! obsolete) {
		    casa.dodwan.util.Logger.log("cache", "load " + key
						+ (cancelled?" (cancelled)":""));
		    cache.load(key, desc);
		}
		else {
		    casa.dodwan.util.Logger.log("cache", "discard " + key
						+ " (obsolete)");
		    if (verbosity.isEnabled())
			System.out.println("  Obsolete");
		}
		
		if (flush || obsolete) {
		    if (verbosity.isEnabled())
			System.out.println("  Flushing " + key);
	
		    int nbFrags = desc.getNbFragments();
		    flush(path, key, nbFrags);
		}
	    }
	    catch (Exception e) {
		if (verbosity.isEnabled())
		    System.out.println("  Failure");
	    }
	}
    }
    

    // ------------------------------------------------------------
    private void flush(File path, String key, int nbFrags) {

	if (verbosity.isEnabled())
		System.out.println("CacheService.flush("
				   + path.getAbsoluteFile() + ", "
				   + key + ")");

	if (cache.contains(key)) {
	    try {
		cache.remove(key);
	    } 
	    catch (Exception e) {}
	    return;
	}

	File file;

	try {
	    File dpath = new File(path, "descriptors");
	    file = new File(dpath, key);
// 	    casa.dodwan.util.Logger.log("cache", "deleting file " + file);
	    file.delete();
	} catch (Exception e) {}

	try {
	    File ppath = new File(path, "payloads");
	    for (int i=0; i < nbFrags; i++) {
		file = new File(ppath, key + "-" + i);
// 	    casa.dodwan.util.Logger.log("cache", "deleting file " + file);
		file.delete();
	    }
	} catch (Exception e) {}
    }
    
    // ------------------------------------------------------------
    /**
     * Sets the verbosity.
     * @param level the verbosity level
     */
    public void setVerbosity(int level) {
	
	switch (level) {
	case 0:
	    verbosity.disable();
	    cache.verbosity.disable();
	    break;
	default:
	    verbosity.enable();
	    cache.verbosity.enable();
	    break;
	}
	
	System.out.println("CacheService.setVerbosity(" + level + ")");
    }
    
    // ------------------------------------------------------------
    /*
     * (non-Javadoc)
     * @see casa.dodwan.util.Suspendable#resume()
     */
    @Override
	public void resume() {
	
	if (verbosity.isEnabled())
	    System.out.println("CacheService.resume()");
	
	if (isEnabled()) 
	    super.resume();
	
	else if (verbosity.isEnabled())
	    System.out.println("  Failure: this service is disabled");
    }
    
    // ------------------------------------------------------------
    /*
     * (non-Javadoc)
     * @see casa.dodwan.util.Suspendable#suspend()
     */
    @Override
	public void suspend() {
	
	if (verbosity.isEnabled())
	    System.out.println("CacheService.suspend()");
	
	if (isEnabled()) {
	    super.suspend();
	}
	else if (verbosity.isEnabled())
	    System.out.println("  Failure: this service is disabled");
    }
    
    // ------------------------------------------------------------
    /**
     * To string.
     * @return String
     */
    public String toString() {

	return "CacheService"
	    + "\n isEnabled:" + String.valueOf(isEnabled())
	    + "\n isSuspended: " + String.valueOf(isSuspended());

    }
    
    // // ------------------------------------------------------------
    // /*
    //  * (non-Javadoc)
    //  * @see casa.dodwan.util.Sink#write(java.lang.Object)
    //  */
    // public void write(Message doc) throws Exception {
	
    // 	if (verbosity.isEnabled()) {
    // 	    String key = doc.getKey();
    // 	    System.out.println("CacheService.write('" + key + "')");
    // 	}
	
    // 	// if (isSuspended()) {
    // 	//     if (verbosity.isEnabled())
    // 	// 	System.out.println("  This service is suspended. Aborting...");
    // 	//     return;
    // 	// }
	
    // 	dispatcher.write(doc);
    // }
    
}
