/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/




package casa.dodwan.config;

import java.io.File;

import casa.dodwan.util.SystemEnvironment;
import casa.dodwan.util.ServiceEnvironment;


/**
 * Provides general purpose variables that influence the behavior of DodWAN.
 * Each of these variable has an initial value specified in the
 * DoDWAN default configuration file or in the user configuration file.
 *
 */


public class DodwanEnvironment extends ServiceEnvironment {


    /**
     * TCP port used by the DoDWAN console for telnet-like
     * control. The DoDWAN console actually uses two TCP ports (that
     * is, <code>console_port</code> and <code>console_port +
     * 1</code>) the first port being reserved for interactive use,
     * and the second port for one-shot commands.
     */
    public int console_port = -1;


    /**
     * @undocumented
     */
    public File directory = null;


    /**
     * Fragmentation threshold (in bytes) for the segmentation and 
     * reassembly of large messages. This threshold applies typically
     * when a message cannot fit in a single UDP datagram.
     */
    public int fragmentation_threshold = -1;

    /**
     * Maximum payload size (in bytes). Messages whose payload is larger than 
     * the specified size cannot be processed by DoDWAN, and are therefore 
     * discarded. If the value of this parameter is set to 0 (zero), then it 
     * means that there is actually no limit to the payload size (use this 
     * possibility with caution!). 
     */
    public int max_payload_size = -1;

    /**
     * @undocumented
     */
    // public File peers_directory = null;


    /**
     * @undocumented
     */
    public String peer_command = null;

 
    /**
     * Cache capacity (in bytes). DoDWAN's cache is maintained in the
     * filesystem, and the payloads of messages are stored in a
     * particular directory. This is a limit on the total size of this
     * directory. The size of message descriptors is not accounted
     * for. If the value of <code>cache_capacity</code> is set to 0,
     * then there is actually no limit to the cache capacity (use this
     * possibility with caution!).
     */
    public int cache_capacity = -1;
    
    /**
     * @undocumented
     * 
     * Specifies how many messages can be proposed simultaneously
     * (i.e. in a single offer) to the same peer.
     */
    public int offer_capacity = 10;

    /**
     * @undocumented
     */
    public long activity_period = 0;

    private static DodwanEnvironment instance_;


    // ------------------------------------------------------------
    /**
     * Obtains the unique instance of class <code>DodwanEnvironment</code>.
     * A new instance is created when this method is called for the first time
     * @return The unique instance
     */
    public static DodwanEnvironment getInstance() {
	if (instance_ == null)
	    instance_ = new DodwanEnvironment();
	return instance_;
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public DodwanEnvironment() {
	
	super("dodwan");

	loadProperties();
    }
    
    // ------------------------------------------------------------
    private void loadProperties() {
	
	// Administration
	console_port =
	    properties.getInteger("dodwan.console_port", -1);
	activity_period = properties.getLong("dodwan.activity_period", 0);
	directory = 
	    new File(properties.getString("dodwan.directory", 
					  SystemEnvironment.varHostPath));
       	// Gossiping
	offer_capacity = 
	    properties.getInteger("dodwan.offer_capacity", 5);
	
	// Cache, segmentation and reassembly
	cache_capacity = 
	    properties.getQuantifiedInteger("dodwan.cache_capacity", 0);
	max_payload_size = 
	    properties.getQuantifiedInteger("dodwan.max_payload_size", 0);
	fragmentation_threshold =
	    properties.getQuantifiedInteger("dodwan.fragmentation_threshold", 
					    1024*1024);

	// Miscellaneous
	// peers_directory = 
	//     new File(properties.getString("dodwan.peers_directory", 
	// 				  "/dev/shm/peers"));
	peer_command = 
	    properties.getString("dodwan.peer_command",
				 "xterm -e ssh");
    }

    // ------------------------------------------------------------
    public void displayProperties() {

	System.out.println("==================================================");
	System.out.println("DoDWAN "+ SystemEnvironment.version + " properties");
	System.out.println("==================================================");
	System.out.println("Gossiping");
	System.out.println("  offer_capacity=" + offer_capacity);
	System.out.println("Cache, segmentation and reassembly");
	System.out.println("  cache_capacity=" + cache_capacity);
	System.out.println("  fragmentation_threshold=" + fragmentation_threshold);
	System.out.println("  max_payload_size=" + max_payload_size);
	System.out.println("Administration");
	System.out.println("  console_port=" + console_port);
	System.out.println("  activity_period=" + activity_period);
	System.out.println("  directory=" + directory);
	System.out.println("Miscellaneous");
	// System.out.println("  peers_directory=" + peers_directory);
	System.out.println("  peer_command=" + peer_command);
	System.out.println("==================================================");
     }

}
