/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/


package casa.dodwan.hub;

import java.io.File;
import java.text.SimpleDateFormat;
import java.util.Locale;
import java.util.Collection;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;

import casa.dodwan.docware.Descriptor;
import casa.dodwan.message.Message;
import casa.dodwan.util.Logger;
import casa.dodwan.util.Processor;
import casa.dodwan.util.DataFormatting;
import casa.dodwan.util.PresenceListener;
import casa.dodwan.net.UDP_Channel;
import casa.dodwan.net.ChannelManager;
import casa.dodwan.hub.SocketRelay;


public class Hub
    extends Thread
    implements Processor<Message>, PresenceListener {

    // ==================================================
    private class Node
	extends Thread {

	public long lastSeen = 0;
	public ServerSocket server = null;
	private String gaddr_ = null;
	private InetSocketAddress gossipingAddr_ = null;
	private Set<SocketRelay> relays_ = new HashSet<SocketRelay>();

	public Node() {
	}

	// --------------------------------------------------
	public void createProxy(String gaddr) {

	    if (gaddr.equals(gaddr_)) {
		// Logger.log("hub", "proxy for gaddr=" + gaddr + " already exists");
		return;
	    }

	    clear();
	    try {
		gaddr_ = gaddr;
		gossipingAddr_ = DataFormatting.toInetSocketAddress(gaddr);
		server = new ServerSocket(0);
		
		Logger.log("hub", "creating proxy socket on port "
			   + server.getLocalPort()
			   + " for peer " + gaddr);
	    }
	    catch (Exception e) {
		System.out.println(e);
		e.printStackTrace();
	    }

	    Thread thread = new Thread(this);
	    this.setName("node-" + gaddr);
	    thread.start();
	}

	// --------------------------------------------------
	public void run() {

	    // System.err.println("Begin " + this.getName());

	    if (server == null)
		return;
	    
	    while(! server.isClosed()) {
		// Logger.log("hub", "waiting for peer on port "
		// 	   + server.getLocalPort());
		try {
		    Socket inSocket = server.accept();

		    Logger.log("hub", "cx accepted from "
		     	       +  DataFormatting.toString((InetSocketAddress)inSocket.getRemoteSocketAddress())
		     	       + ", connecting to "
		     	       +  gaddr_);
		    Socket outSocket = new Socket(gossipingAddr_.getAddress(),
						  gossipingAddr_.getPort());
		    SocketRelay relay = new SocketRelay(inSocket, outSocket, environment);
		    relays_.add(relay);
		}
		catch (Exception e) {
		    // System.out.println(e);
		    // e.printStackTrace();
		}
	    }
	    Logger.log("hub", "proxy socket port "
		       + server.getLocalPort()
		       + " closed");

	    // System.err.println("Terminate " + this.getName());
	}

	// --------------------------------------------------
	public void clear() {

	    if (server == null)
		return;
	    
	    Logger.log("hub", "clearing proxy socket listening on " + server.getLocalPort());

	    try {
		server.close();
	    }
	    catch(Exception e) {
	    }
	    
	    for (SocketRelay relay: relays_) 
		relay.close();
	    relays_.clear();
	}
    }

    // ==================================================
    public Environment environment;
    public long latency = 60;
    public long period  = 10;
    private boolean dirMode_ = false;

    private Map<String,Node> nodes_ = new HashMap<String,Node>();

    private UDP_Channel channel_;

    private ChannelManager manager_;

    private InetAddress localAddr_;
    
    // -------------------------------------------------------------------
    public Hub(int port,
	       Environment env,
	       boolean dir_mode)
	throws Exception {

	resetLogger();

	// Test DoDWAN/aDTN
	localAddr_ = casa.dodwan.net.browseInterfaces.getInet4Address();
	// localAddr_ = InetAddress.getByName("192.168.196.11");
	
	Logger.log("hub", "local=" + DataFormatting.toString(localAddr_)
		   + " for TCP-based gossiping (if any)");


	manager_ = new ChannelManager(this, this);
	
	channel_ = new UDP_Channel(new InetSocketAddress("::", port),
				   null, null,
				   0, false, manager_);
	channel_.passthrough = true;
	manager_.addChannel(channel_);

	dirMode_ = dir_mode;

	environment = env;

	channel_.open();
    }

    // -------------------------------------------------------------------
    /**
     *  @undocumented
     *  Starts logging Hub events in a new file. 
     */
    public static void resetLogger() {
	
	// System.out.println("Hub.resetLogger");

	SimpleDateFormat dateFormat = 
	    new SimpleDateFormat("yyyy-MM-dd-HH-mm", Locale.ENGLISH);
	long now = System.currentTimeMillis();
	String fname = "hub"
	    // + "_" + dateFormat.format(now) 
	+ ".log";
	    
	// File logFile = new File(dir, fname);
	File logFile = new File(fname);

	Logger.enable(null, true, "hub");

	// System.out.println("Now logging in file "
	// 		   + logFile);
    }

    // ============================================================
    // MANAGEMENT OF CLIENT NODES
    // ============================================================

    private void purgeOldNodes() {

	// Logger.log("hub", "purging old nodes (if any...)");
  	long now = System.currentTimeMillis();
	
	// Let us remove nodes we haven't heard about for a long time
	synchronized(environment) {
	    Set<String> oldNodes = new HashSet<String>();
	    for (String id: nodes_.keySet()) {
		Node node = nodes_.get(id);
		long elapsed = (now - node.lastSeen) / 1000;
		if (elapsed > latency) {
		    oldNodes.add(id);
		    Logger.log("hub", " purge " + id
			       + " (elapsed=" + elapsed
			       + " sec)");
		}
	    }
	    for (String id: oldNodes)
		removeNode(id);
	}
    }

    // ------------------------------------------------------------
    private void removeNode(String id) {

	Node node = nodes_.get(id);
	if (node != null) {
	    node.clear();
	    nodes_.remove(id);
	}
	environment.deleteNode(id);
    }
    
    // ------------------------------------------------------------
    public void run() {

	try {
	    while(true) {

		sleep(period * 1000);
		purgeOldNodes();
	    }
	}
	catch(Exception e) {
	    System.out.println("Hub.run(): " + e);
	}
    }

    // ============================================================
    // Implementation of the Processor<Message> interface
    // ============================================================

    private void processHelloBeacon(String sdr,
				    boolean dup,
				    Descriptor desc) {
	
	// Processing a "hello" beacon

	String baddr = desc.getAttribute("_baddr");
	String gaddr = desc.getAttribute("_gaddr");
	
	Logger.log("hub", "< sdr=" + sdr
		   + " (hello) "
		   + (baddr==null?"":baddr)
		   + (gaddr==null?"":gaddr));

	if (dup)
	    environment.setNode(sdr, true);
	// else
	//   environment.addNode(sdr, true);

	Node node = nodes_.get(sdr);
	if (node == null) {
	    node = new Node();
	    nodes_.put(sdr, node);
	}

	// If a gossiping address is specified in that beacon, let us
	// try to create a socket proxy for TCP connections to that
	// address
	if (gaddr != null) 
	    node.createProxy(gaddr);
	    

	// Recording the date of the last beacon received from
	// that peer
	long now = System.currentTimeMillis();
	node.lastSeen = now;
	
	if (dirMode_) {
	    // If the hub is running in directory mode, then
	    // any "hello" announce it forwards should contain
	    // an '_addr' attribute that specifies the address
	    // and port number of the actual sender of that
	    // announce
	    InetSocketAddress beaconingAddr = channel_.remoteAddress(sdr);
	    if (baddr != null)
		desc.setAttribute("_baddr",
				  DataFormatting.toString(beaconingAddr));
	}
	else
	    desc.removeAttribute("_baddr");
	
	// If the beacon contains a "_gaddr" attribute, then
	// the value of this attribute should be modified so
	// all receiving peers connect to the hub instead
	if (gaddr != null)
	    desc.setAttribute("_gaddr", DataFormatting.toString(localAddr_)
			      + ":"
			      + node.server.getLocalPort());
    }

    // ------------------------------------------------------------
    public void process(Message message) throws Exception {

	Descriptor desc = message.getDescriptor();
	
	String key   = desc.getKey();
	String sdr   = desc.getAttribute("_sdr");
	String rcv   = desc.getAttribute("_rcv");
	String btype = desc.getAttribute("_btype");
	
	if (sdr == null) {
	    Logger.log("hub", "warning: sdr=null (msg discarded)");
	    return;
	}

	synchronized(environment) {
	    boolean hello = false;
	    boolean bye   = false;
	    boolean dup   = false;
	    if (btype != null) {
		hello = btype.equals("hello");
		if (hello)
		    dup = nodes_.containsKey(sdr);
		bye   = btype.equals("bye");
	    }
	    
	    Logger.log("hub", "< sdr=" + sdr
		       + ",rcv=" + rcv
		       + (hello?" (hello)":"")
		       + (dup?" [dup]":"")
		       + (bye?" (bye)":""));
	    
	    if (hello)
		processHelloBeacon(sdr, dup, desc);

	    if (rcv != null) {
		// Sending message to a single receiver, as requested by
		// sender
		if (environment.arePeers(sdr, rcv)) {
		    Logger.log("hub", "  > sdr=" + sdr
			       + ",rcv=" + rcv);
		    channel_.write(message);
		}
	    }
	    else {
		// Forwarding message to all peers of the sender
		Collection<String> peers = environment.getPeers(sdr);
		Set<String> targets = new HashSet<String>();
		for (String id: peers) 
		    if (environment.isOnline(id)) 
			targets.add(id);
		if (! targets.isEmpty()) {
		    Logger.log("hub", "  > sdr=" + sdr
			       + ",rcv=" 
			       + DataFormatting.toString(peers, ","));
		    for (String id: targets) {
			desc.setAttribute("_brcv", id);
			channel_.write(message);
		    }
		}
	    }
	    
	    if (bye) {
		// environment.setNode(sdr, false);
		// FG: should offline nodes be removed from the
		// simulation?
		// environment.deleteNode(sdr);
		// nodes_.remove(sdr);
	    }
	}
    }

    // ------------------------------------------------------------
    // Implementation of the PresenceListener interface
    // ------------------------------------------------------------
    public void addPeer(String id) {
    	Logger.log("hub", "Hub.addPeer(" + id + ")");
	environment.addNode(id, true);
    }

    public void removePeer(String id) {
    	Logger.log("hub", "Hub.removePeer(" + id + ")");
	// FG: permutation deux lignes ci-dessous pour test passage
	// offline avec mobsim en mode visuel
	environment.setNode(id, false);
	// removeNode(id);
    }

    public void updatePeer(String id) {}

    public void clearPeers() {}
}
