/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.message;

import java.util.Date;
import java.io.File;
import java.io.PrintStream;
import java.io.InputStream;
import java.io.FileInputStream;

import casa.dodwan.docware.Descriptor;
import casa.dodwan.docware.Payload;
import casa.dodwan.cache.Cache;
import casa.dodwan.util.Compressor;
import casa.dodwan.util.Sink;
import casa.dodwan.util.StringGenerator;
import casa.dodwan.util.Verbosity;

public class MessageSAR {
    
    public static Verbosity verbosity = new Verbosity();
    
    // ------------------------------------------------------------
    public static void publish(Descriptor descriptor, 
			       byte[] buffer,
			       int threshold,
			       Cache cache)
	throws Exception {
	
	if (verbosity.isEnabled()) {
	    System.out.println("MessageSAR.publish()");
	}
	
	int size = (int) buffer.length;

	String docId = descriptor.getDocumentId();
	if (docId == null)
	    docId = StringGenerator.getGenerator().newString();
	
	casa.dodwan.util.Logger.log("sar", "publish(id="
				    + docId 
				    + ",buffer["
				    + buffer.length
				    + "])");

	int fragmentSize = threshold;
	int nbFrags = (size + fragmentSize - 1)
	    / fragmentSize;

	// If size is 0, then let's cheat and make sure and empty byte
	// array is sent anyway
	if (nbFrags == 0)
	    nbFrags = 1;

	// descriptor.setAttribute("_cancelled", "false");
	descriptor.setDate(new Date(System.currentTimeMillis()));
	descriptor.setDocumentId(docId);
	descriptor.setNbFragments(nbFrags);
	descriptor.setAttribute("_weight", String.valueOf(size));
	cache.addDescriptor(descriptor);

	int curLength = 0;
	for (int fragId = 0; fragId < nbFrags; fragId++) {
	    
	    // Message fragment = new Message(new Descriptor(descriptor));
	    // Descriptor desc = fragment.getDescriptor();
	    // desc.setDate(new Date(System.currentTimeMillis()));
	    // desc.setDocumentId(docId + "-" + fragId);
	    // desc.setFragmentId(fragId);
	    // desc.setNbFragments(nbFrags);
	    // if (fragId == 0)
	    // 	desc.setAttribute("_weight", 
	    // 			  String.valueOf(size));
	    
	    if (fragId < nbFrags - 1)
		curLength = fragmentSize;
	    else
		curLength = size % fragmentSize;
	    
	    byte[] payload = new byte[curLength];
	    
	    System.arraycopy(buffer, fragId * fragmentSize, payload,
			     0, curLength);
	    
	    // fragment.setPayload(payload);
	    // cache.add(fragment);

	    cache.addPayload(docId, fragId, payload);
	}
    }

    // ------------------------------------------------------------
    public static void publish(Descriptor descriptor, 
			       File file,
			       int threshold,
			       Cache cache)
	throws Exception {
	
	if (verbosity.isEnabled()) {
	    System.out.println("MessageSAR.publish()");
	}
	
	int size = (int) file.length();
	InputStream in = new FileInputStream(file);

	String docId = descriptor.getDocumentId();
	if (docId == null)
	    docId = StringGenerator.getGenerator().newString();
	
	casa.dodwan.util.Logger.log("sar", "publish(id="
				    + docId
				    + ",file="
				    + file 
				    + "["
				    + size + "])");

	int fragmentSize = threshold;
	int nbFrags = (size + fragmentSize - 1)
	    / fragmentSize;
	
	// If size is 0, then let's cheat and make sure and empty byte
	// array is sent anyway
	if (nbFrags == 0)
	    nbFrags = 1;

	// descriptor.setAttribute("_cancelled", "false");
	descriptor.setDate(new Date(System.currentTimeMillis()));
	descriptor.setDocumentId(docId);
	descriptor.setNbFragments(nbFrags);
	descriptor.setAttribute("_weight", String.valueOf(size));
	cache.addDescriptor(descriptor);

	int curLength = 0;
	for (int fragId = 0; fragId < nbFrags; fragId++) {
	    
	    // Message fragment = new Message(new Descriptor(descriptor));
	    // Descriptor desc = fragment.getDescriptor();
	    // desc.setDate(new Date(System.currentTimeMillis()));
	    // desc.setDocumentId(docId + "-" + fragId);
	    // desc.setFragmentId(fragId);
	    // desc.setNbFragments(nbFrags);
	    // if (fragId == 0)
	    // 	desc.setAttribute("_weight", 
	    // 			  String.valueOf(size));
	    
	    if (fragId < nbFrags - 1)
		curLength = fragmentSize;
	    else
		curLength = size % fragmentSize;
	    
	    byte[] payload = new byte[curLength];
	    int nread = in.read(payload, 0, curLength);
	    
	    // fragment.setPayload(payload);
	    // cache.add(fragment);
	    cache.addPayload(docId, fragId, payload);
	}
	in.close();
    }

    // ------------------------------------------------------------
    public static byte[] getAsBytes(String key, Cache cache)
	throws Exception {

	casa.dodwan.util.Logger.log("sar", "getAsBytes(id=" + key + ")");

	if (verbosity.isEnabled())
	    System.out.print("PubSubService.getAsBytes()");
	
	// Extracting the number of fragments and payload size from
	// the descriptor
	Descriptor desc = cache.getDescriptor(key);
	if (desc == null)
	    throw new Exception("Descriptor " + key + " is missing");

	int nbFrags = desc.getNbFragments();
	int size = Integer.parseInt(desc.getAttribute("_weight"));
	byte[] buffer = new byte[size];

	// Extracting all fragments of the payload from cache
	int offset = 0;
	for (int cnt = 0; cnt<nbFrags; cnt++) {
	    String id = key + "-" + cnt;
	    Payload payload = cache.getPayload(id);
	    if (payload == null)
		throw  new Exception("Fragment " + id + " is missing");
	    byte[] fragment = payload.getBytes();
	    if (offset + fragment.length > buffer.length)
		throw  new Exception("Buffer overflow with fragment " + id);
	    System.arraycopy(fragment, 0, buffer,
			     offset, fragment.length);
	    offset += fragment.length;
	}

	return buffer;
    }

    // ------------------------------------------------------------
    public static void getAsFile(String key, File file, Cache cache)
	throws Exception {

	casa.dodwan.util.Logger.log("sar", "getAsFile(id=" + key 
			     + ",file=" + file + ")");

	if (verbosity.isEnabled())
	    System.out.print("PubSubService.getAsFile()");
	
	PrintStream out = new PrintStream(file);

	// Extracting the number of fragments from the descriptor
	Descriptor desc = cache.getDescriptor(key);
	if (desc == null)
	    throw new Exception("Descriptor " + key + " is missing");

	int nbFrags = desc.getNbFragments();

	// Extracting all fragments of the payload from cache
	for (int cnt = 0; cnt<nbFrags; cnt++) {
	    String id = key + "-" + cnt;
	    Payload payload = cache.getPayload(id);
	    if (payload == null)
		throw  new Exception("Fragment " + id + " is missing");
	    byte[] fragment = payload.getBytes();
	    out.write(fragment);
	}

	out.flush();
	out.close();
    }

}
