/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.net;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.SocketAddress;

import casa.dodwan.util.Bufferizer;
import casa.dodwan.util.DataBufferizer;
import casa.dodwan.util.Logger;
import casa.dodwan.util.Socket;
import casa.dodwan.util.Verbosity;

public class InetSocket<DOCTYPE> {

	public static final int MAX_BUFFER_SIZE = 65507;

	public InetSocketAddress echoAddress = null;
	public long lastEcho = 0;

	protected Bufferizer<DOCTYPE> bufferizer_;

	protected int bufSize_ = MAX_BUFFER_SIZE;

	protected InetSocketAddress localSocketAddress_;

	protected DatagramPacket outPacket_;

	protected DatagramSocket socket_;

	protected InetAddress lastSndAddress_ = null;
	// protected long lastSndTime_ = 0;
	protected InetSocketAddress lastRcvAddress_ = null;
	// protected long lastRcvTime_ = 0;

	protected static byte cookie_ = 31;

	public Verbosity verbosity = new Verbosity();

	protected InetSocket() {
	};

	// ------------------------------------------------------------
	public InetSocket(InetSocketAddress localSocketAddress,
			  Bufferizer<DOCTYPE> bufferizer) throws Exception {		
	    
	    //  	    verbosity.enable();
	    
	    if (verbosity.isEnabled())
		System.out.println("InetSocket(" + localSocketAddress + ", "
				   + bufferizer + ")");
	    
	    init(localSocketAddress, bufferizer);

	    long now = System.currentTimeMillis();
	    lastEcho = now;
	}

	// ------------------------------------------------------------
	public InetSocket(Bufferizer<DOCTYPE> bufferizer) throws Exception {		
	    
	    //  	    verbosity.enable();
	    
	    if (verbosity.isEnabled())
		System.out.println("InetSocket(" + bufferizer + ")");
	    
	    init(bufferizer);
	}

	// ------------------------------------------------------------
	public void close() throws Exception {

	    socket_.close();
	}

	// ------------------------------------------------------------
	private void init(InetSocketAddress localSocketAddress,
			  Bufferizer<DOCTYPE> bufferizer) throws Exception {
	    
	    localSocketAddress_ = localSocketAddress;
	    bufferizer_ = bufferizer;
	    
	    socket_ = new DatagramSocket(localSocketAddress_);
	    
	    Logger.log("socket", "local=" + localSocketAddress_);
	    Logger.log("socket", "cookie=" + cookie_);
	    Logger.log("socket", "rcvBufferSize=" + socket_.getReceiveBufferSize());
	}

	// ------------------------------------------------------------
	private void init(Bufferizer<DOCTYPE> bufferizer) throws Exception {
	    
	    bufferizer_ = bufferizer;

	    socket_ = new DatagramSocket();
	    localSocketAddress_ = (InetSocketAddress)socket_.getLocalSocketAddress();
	    
	    Logger.log("socket", "local=" + localSocketAddress_);
	    Logger.log("socket", "cookie=" + cookie_
		       + ",rcvBufferSize=" + socket_.getReceiveBufferSize());
	}

	// ------------------------------------------------------------
	public InetSocketAddress localSocketAddress() {
	    return localSocketAddress_;
	}

	// ------------------------------------------------------------
	public DOCTYPE read() throws Exception {

	    byte[] buffer = this.receiveBytes();
	    
	    DOCTYPE value = null;
	    try {
		value = bufferizer_.fromBuffer(buffer, 0, buffer.length);
	    }
	    catch (Exception e) {
		System.out.println("InetSocket.read(): extraction from buffer failed");
		throw e;
	    }
	    
	    if (verbosity.isEnabled())
		System.out.println("InetSocket.read('" + value + "')");
	    
	    return value;
	}
	
	// ------------------------------------------------------------
	/*
	 * Receives a message through this <code>InetSocket</code>,
	 * and returns the content of this message as a byte array.
	 * 
	 * @return the byte array that contains the message received.
	 */
	protected byte[] receiveBytes() throws IOException {

	    DatagramPacket inPacket =
		new DatagramPacket(new byte[bufSize_], bufSize_);
	    
	    // Receive packets until one is not an echo packet
	    // (sent by ourself)
	    boolean echo;
	    do { 
		socket_.receive(inPacket);
		
		lastRcvAddress_ = (InetSocketAddress)inPacket.getSocketAddress();

		// Discard echo messages
		echo = lastRcvAddress_.equals(echoAddress);
		if (echo) {
		    long now = System.currentTimeMillis();
		     // Logger.log("socket", 
		     // 	       "<X echo ("
		     // 	       + (now - lastEcho)
		     // 	       + ")");
		    lastEcho = now;
		}
	    } while (echo == true);
	    // lastRcvTime_ = System.currentTimeMillis();

	    int nbBytes = inPacket.getLength();
	    int offset = inPacket.getOffset();
		
	    // Logger.log("socket", "<["
	    //  	       + nbBytes + "] " 
	    //  	       + inPacket.getAddress());
		
	    byte[] data = inPacket.getData();
		
	    // Checking magic cookie in payload
	    if (data[0] != cookie_) {
		System.out.println("InetSocket.receiveBytes(): invalid cookie rcvd from "
				   + inPacket.getSocketAddress());
		throw new IOException("Invalid cookie");
	    }
	    
	    byte[] result = new byte[nbBytes-1];
	    DataBufferizer.copyBuffer(data, offset+1, result, 0, nbBytes-1);
	    
	    if (verbosity.isEnabled())
		System.out.println("InetSocket.receiveBytes() from "
				   + lastRcvAddress_);
	    // inPacket.getSocketAddress());
	    return result;
	}
	
    // ------------------------------------------------------------
    public InetSocketAddress lastRcvAddress() {
	    return lastRcvAddress_;
	}

	// ------------------------------------------------------------
	public InetAddress lastSndAddress() {
	    return lastSndAddress_;
	}

	// ------------------------------------------------------------
	/*
	 * Sends a fragment of a byte array through this
	 * <code>InetSocket</code>.
	 * 
	 * @param addr the target address. @param port the target
	 * port. @param buffer the byte array whose content must be
	 * sent through this <code>InetSocket</code>. @param offset
	 * the offset of the first byte in the buffer that must be
	 * sent through the socket. @param length the number of bytes
	 * that must be sent through the socket.
	 */
	protected void sendBytes(InetAddress addr, int port,
				 byte[] buffer, int offset, int length)
			throws IOException {

	    if (verbosity.isEnabled()) 
		System.out.println("InetSocket.sendBytes(" + addr
				   + ":" + port + ", "
				   + buffer + ", " + offset
				   + ", " + length + ")");
	    
	    try {
		outPacket_ = new DatagramPacket(new byte[bufSize_], bufSize_);
		outPacket_.setAddress(addr);
		outPacket_.setPort(port);
		byte[] data = outPacket_.getData();

		// Inserting magic cookie in buffer
		data[0] = cookie_;

		// Copying data in buffer after the magic cookie
		DataBufferizer.copyBuffer(buffer, offset, data, 1, length);

		// Sending packet
		outPacket_.setLength(length+1);
		socket_.send(outPacket_);
	
		lastSndAddress_ = outPacket_.getAddress();
		// lastSndTime_ = System.currentTimeMillis();

		 // Logger.log("socket", ">["
		 // 	    + outPacket_.getLength() + "] " 
		 // 	    + outPacket_.getAddress() + ":" 
		 // 	    + outPacket_.getPort());
		
	    } catch (IOException e) {
		if (verbosity.isEnabled())
		    System.out.println("InetSocket: could not send packet");
		throw e;
	    }
	    
	    if (verbosity.isEnabled()) 
		System.out.println("InetSocket... packet= " + outPacket_);
	}
	
	// ------------------------------------------------------------
	public int getTTL(){
		return 0;
	}
	
	// ------------------------------------------------------------
	public void setTTL(int ttl){}	
	
	// ------------------------------------------------------------
	@Override
	public String toString() {
	    String result = "InetSocket(" + localSocketAddress_ + ")";
	    return result;
	}

	// ------------------------------------------------------------
	public void write(InetAddress remoteAddress,
			  int remotePort,
			  DOCTYPE value) throws Exception {
		
	    if (verbosity.isEnabled())
		System.out.println("InetSocket.write(" + remoteAddress
				   + "," + remotePort
				   + "," + value + ")");
	    
	    byte[] buffer = new byte[MAX_BUFFER_SIZE];
	    int length = bufferizer_.toBuffer(value, buffer, 0);
	    
	    try {
		this.sendBytes(remoteAddress,
			       remotePort,
			       buffer, 0, length);
	    } catch (Exception e) {
		if (verbosity.isEnabled())
		    System.out.println("  Failed\n" + e);
	    }
	}
	
	// ------------------------------------------------------------
	public void write(InetSocketAddress remoteSocketAddress,
			  DOCTYPE value) throws Exception {

	    write(remoteSocketAddress.getAddress(), 
		  remoteSocketAddress.getPort(),
		  value);
	}
}
