/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/


package casa.dodwan.net;

import java.io.IOException;
import java.net.SocketAddress;
import java.net.InetSocketAddress;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.DatagramSocket;
import java.net.MulticastSocket;
import java.net.DatagramPacket;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.Iterator;
import java.net.SocketTimeoutException;
import java.util.HashMap;
import java.io.PrintStream;

import casa.dodwan.util.ArgumentNotFoundException;
import casa.dodwan.util.ArgumentParsing;
import casa.dodwan.util.Logger;
import casa.dodwan.util.SystemEnvironment;

import casa.dodwan.util.DataBufferizer;
import casa.dodwan.util.DataFormatting;
import casa.dodwan.docware.Descriptor;
import casa.dodwan.message.Message;
import casa.dodwan.message.MessageBufferizer;
import casa.dodwan.config.DodwanEnvironment;

public class UDP_Channel 
    extends Thread 
    implements Channel {

    // ==================================================
    private class Peer {

	InetSocketAddress address;
	long lastRcvd;
	long lastSent;

	public Peer(InetSocketAddress addr) {
	    address = addr;
	}
    }
    
    // ==================================================
    
    public static final int MAX_BUFFER_SIZE = 65507;

    public static byte COOKIE = 31;

    private DatagramSocket socket_ = null;

    private InetSocketAddress localSocketAddr_  = null;
    private InetSocketAddress remoteSocketAddr_ = null;
    private InetAddress       intfAddr_         = null;
    
    private MessageBufferizer out_bufferizer_ = 
	new MessageBufferizer();
    private MessageBufferizer in_bufferizer_ = 
	new MessageBufferizer();

    // private long lastSndTime_ = 0;
    // private long lastRcvTime_ = 0;
    private InetSocketAddress lastRcvAddr_ = null;
    
    private ChannelManager manager_ = null;

    private BeaconSender sender_ = null;
    private boolean use_tcp = false;
    private TCP_Server tcp_server_ = null;
    private InetSocketAddress tcpSocketAddr_ = null;
    
    private boolean isOpen_ = false;
    private String hostId_  = null;
    
    private Map<String,Peer> peers_ = 
	new HashMap<String,Peer>();

    private long period_ = 0;
    private long latency_ = 0;
    
    public boolean passthrough = false;
    
    // ============================================================

    public static boolean checkArguments(String localAddr,
					 int localPort,
					 String remoteAddr,
					 int remotePort,
					 String intfAddr,
					 PrintStream out) {

	out.println("Checking socket parameters: local="
		    + localAddr + "/" + localPort
		    + ",remote="
		    + remoteAddr + "/" + remotePort
		    + (intfAddr == null?"":",iface=" + intfAddr));

 	boolean check = true;
	try {
	    InetAddress addr = InetAddress.getByName(localAddr);
	}
	catch (Exception e) {
	    out.println(localAddr + " is not a valid address");
	    check = false;
	}
	
	if ((localPort < 0) || (localPort > 65535)) {
	    out.println(localPort + " is not a valid port number");
	    check = false;
	}
	
	if ((remotePort < 0) || (remotePort > 65535)) {
	    out.println(remotePort + " is not a valid port number");
	    check = false;
	}

	try {
	    InetAddress addr = InetAddress.getByName(remoteAddr);
	    if (addr.isMulticastAddress() && (localPort != remotePort)) {
		    out.println("Local and remote port numbers should be similar with a multicast address");
		    check = false;
		}
		}
	catch (Exception e) {
	    out.println(remoteAddr + " is not a valid address");
	    check = false;
	}

	InetAddress interfaceAddr = null;
	if (intfAddr != null) {
	    try {
		interfaceAddr = InetAddress.getByName(intfAddr);
		NetworkInterface nif =
		    NetworkInterface.getByInetAddress(interfaceAddr);
		if (nif == null) {
		    out.println(intfAddr + " is not the address of a local interface");
		    check = false;
		}
	    }
	    catch (Exception e) {
		out.println(intfAddr + " is not a valid address");
		check = false;
	    }
	}
	
	if (check)
	    out.println("  OK");
	else
	    out.println("  Invalid parameters");

	return check;
    }

    // ------------------------------------------------------------
    public static String getSyntax() {

	return "[-la localAddr] [-lp localPort] <-ra remoteAddr> <-rp remotePort> [-ia interfAddr] [-p period] [-use_tcp true|false]";
    }

    // ------------------------------------------------------------
    public UDP_Channel(String[] arguments,
		       ChannelManager manager)
	throws Exception {

	manager_ = manager;
	
	String localAddr = "::";
	String remoteAddr = null;
    	int localPort = 0;
    	int remotePort = 0;
	String intfAddr = null;
	long period = 15;
	boolean use_tcp = false;
	
	try {
	    localAddr = ArgumentParsing.searchString("-la", arguments);
	} catch (ArgumentNotFoundException e) {} 
	remoteAddr = ArgumentParsing.searchString("-ra", arguments);
	try {
	    intfAddr = ArgumentParsing.searchString("-ia", arguments);
	} catch (ArgumentNotFoundException e) {} 
	try {
	    localPort = ArgumentParsing.searchInt("-lp", arguments);
	} catch (ArgumentNotFoundException e) {} 
	remotePort = ArgumentParsing.searchInt("-rp", arguments);
	try {
	    period = ArgumentParsing.searchLong("-p", arguments);
	} catch (ArgumentNotFoundException e) {} 
	try {
	    use_tcp = ArgumentParsing.searchBoolean("-use_tcp", arguments);
	} catch (ArgumentNotFoundException e) {} 

	// if (localAddr == null)
	//     localAddr = DataFormatting.toString(casa.dodwan.net.browseInterfaces.getInet4Address());

	init(new InetSocketAddress(localAddr, localPort),
	     new InetSocketAddress(remoteAddr, remotePort),
	     (intfAddr==null?null:InetAddress.getByName(intfAddr)),
	     period,
	     use_tcp);
    }
    
    // ------------------------------------------------------------
    public UDP_Channel(InetSocketAddress localSocketAddr,
		       InetSocketAddress remoteSocketAddr,
		       InetAddress intfAddr,
		       long period,
		       boolean use_tcp,
		       ChannelManager manager)
	throws Exception {

	manager_ = manager;
	
	init(localSocketAddr,
	     remoteSocketAddr,
	     intfAddr,
	     period,
	     use_tcp);
    }

    // ------------------------------------------------------------
    private void init(InetSocketAddress localSocketAddr,
		      InetSocketAddress remoteSocketAddr,
		      InetAddress intfAddr,
		      long period,
		      boolean use_tcp) 
	throws Exception {

	localSocketAddr_ = localSocketAddr;
	remoteSocketAddr_ = remoteSocketAddr;
	intfAddr_ = intfAddr;

	UDP_Channel result = null;

	boolean useMulticast =
	    ((remoteSocketAddr_ != null)
	     && remoteSocketAddr_.getAddress().isMulticastAddress());
	
	try {	
	    if (useMulticast) {
		// Opening a multicast socket
		initMulticast(localSocketAddr_,
			      remoteSocketAddr_,
			      intfAddr_,
			      period,
			      use_tcp);
	    }
	    else 
		initUnicast(localSocketAddr_,
			    remoteSocketAddr_,
			    period,
			    use_tcp);
	} catch (Exception e) {
	    Logger.log("udp_chan", "getting socket failed");
	    System.out.println("   Failed");
	    throw e;
	}
	// System.out.println("Socket=" + socket);
    }

    // ------------------------------------------------------------
    private void initUnicast(InetSocketAddress localSocketAddr,
			     InetSocketAddress remoteSocketAddr,
			     long period,
			     boolean use_tcp) 
	throws Exception {

	hostId_ = SystemEnvironment.host;
	
	remoteSocketAddr_ = remoteSocketAddr;
	InetAddress localAddr = localSocketAddr.getAddress();
	
	socket_ = new DatagramSocket(localSocketAddr);
	localSocketAddr_  = (InetSocketAddress)socket_.getLocalSocketAddress();
	
	prepareBeaconSender(period);

	Logger.log("udp_chan", "create " + this);

	if (use_tcp) 
	    startTCP_Server(localAddr, 0);
	    // startTCP_Server(localPort);
    }
    
    // ------------------------------------------------------------
    private void initMulticast(InetSocketAddress localSocketAddr,
			       InetSocketAddress remoteSocketAddr,
			       InetAddress interfaceAddr,
			       long period,
			       boolean use_tcp) 
	throws Exception {

	hostId_ = SystemEnvironment.host;

	InetAddress mcastAddr = remoteSocketAddr.getAddress();
	int port = remoteSocketAddr.getPort();
	
	if (! mcastAddr.isMulticastAddress())
	    throw new Exception("Invalid multicast address");

	InetAddress localAddr = localSocketAddr.getAddress();
	localSocketAddr_  = new InetSocketAddress("::", port);
	remoteSocketAddr_ = remoteSocketAddr;
	
	if (interfaceAddr != null) {
	    NetworkInterface nif = 
		NetworkInterface.getByInetAddress(interfaceAddr);
	    if (nif == null) {
		throw new Exception("Invalid interface address");
	    }
	}

	MulticastSocket msocket =
	    new MulticastSocket(port);
	// Loopback mode must be enabled, otherwise multicast packets
	// won't be received through that socket
	msocket.setLoopbackMode(true);
	socket_ = msocket;

	Logger.log("udp_chan", "create(local="
		   + DataFormatting.toString(localSocketAddr_)
		   + ",remote="
		   + DataFormatting.toString(remoteSocketAddr_) + ")");

	if (interfaceAddr == null) {
	    Logger.log("udp_chan", "join group " + mcastAddr.getHostAddress()
		       + " (iface=dflt)");
	    msocket.joinGroup(mcastAddr);
	}
	else {
	    NetworkInterface nif = 
		NetworkInterface.getByInetAddress(interfaceAddr);
	    if (nif == null) {
		throw new Exception("Invalid interface address");
	    }
	    Logger.log("udp_chan", "join group " + mcastAddr.getHostAddress()
		       + " (iface=" + interfaceAddr.getHostAddress() + ")");
	    msocket.joinGroup(new InetSocketAddress(mcastAddr, 0), nif);
	}

	prepareBeaconSender(period);

    	if (use_tcp) 
	    startTCP_Server(localAddr, 0);
	    // startTCP_Server(localSocketAddr.getPort());
    }
	
    // ------------------------------------------------------------
    private void startTCP_Server(InetAddress localAddr, int port) {

	try {
	    tcp_server_ = new TCP_Server(port, manager_);
	    tcp_server_.start();
	    use_tcp = true;
	}
	catch (Exception e) {
	    Logger.log("tcp_server", "could not start TCP server");
	}

	// Cannot use a wildcard address for TCP-based gossiping. If
	// the UDP_Channel is not bound to a non-wildcard address,
	// then let us try to discover a valid IPv4 address to be used
	// for gossiping.
	if (localAddr == null) 
	    localAddr = localSocketAddr_.getAddress();
	if (localAddr.isAnyLocalAddress()) 
	    localAddr = casa.dodwan.net.browseInterfaces.getInet4Address();
	
	Logger.log("tcp_server", "will use "
		   + DataFormatting.toString(localAddr)
		   + " as a local address for TCP gossiping");
	InetSocketAddress serverAddress =
	    new InetSocketAddress(localAddr,
				  tcp_server_.getPort());
	sender_.setAttribute("_gaddr",
			     DataFormatting.toString(serverAddress));

    }

    // ------------------------------------------------------------
    private void prepareBeaconSender(long period) {

	period_ = period;
	latency_ = (long)(2.1 * (float)period_);
	  
	if (period_ > 0) {
	    Logger.log("udp_chan", "beaconing period: "
		       + period + " s");
	    sender_ = new BeaconSender(this, period_ * 1000);
	    sender_.enable();
	}
    }
    
    // ------------------------------------------------------------
    public void setRemoteAddress(InetSocketAddress defaultRemoteAddr) 
	throws Exception {

	remoteSocketAddr_ = defaultRemoteAddr;
    }

    // ------------------------------------------------------------
    public InetSocketAddress remoteAddress() {

	return remoteSocketAddr_;
    }

    // ------------------------------------------------------------
    public InetSocketAddress remoteAddress(String id) {

	Peer peer = peers_.get(id);
	if (peer == null)
	    return null;
	else
	    return peer.address;
    }
    
    // ------------------------------------------------------------
    public void setPeer(String peerId, 
			InetSocketAddress addr) {
	
	Peer peer = peers_.get(peerId);
	
	if (peer == null) {
	    Logger.log("udp_chan", "register peer=" + peerId
		       + ",addr=" + DataFormatting.toString(addr));
	    peer = new Peer(addr);
	    peer.lastRcvd = System.currentTimeMillis();
	    peers_.put(peerId, peer);
	}
	else {
	    if (! peer.address.equals(addr)) {
		Logger.log("udp_chan", "change peer=" + peerId
			   + ",addr=" + DataFormatting.toString(addr));
		peer.address = addr;
	    }
	}
    }
    
    // ------------------------------------------------------------
    public Set<String> getPeers() {
	
	return peers_.keySet();
    }

    // ------------------------------------------------------------
    public InetSocketAddress localAddress() {

	return (InetSocketAddress)socket_.getLocalSocketAddress();
    }

    // ------------------------------------------------------------
    public int localPort() {

	return socket_.getLocalPort();
    }
    
    // ------------------------------------------------------------
    private void processHelloAnnounce(String sdr,
				      String rcv,
				      String baddr,
				      String gaddr,
				      Message msg,
				      InetSocketAddress sockAddr) {

	// Let us check if '_sdr' is already registered as a channel
	boolean newPeer = (! peers_.containsKey(sdr));
	
	Logger.log("udp_chan", "< (hello) sdr=" + sdr
    	  	   + (rcv==null?"":",rcv=" + rcv)
    	  	   + (baddr==null?"":",baddr=" + baddr)
    	  	   + (gaddr==null?"":",gaddr=" + gaddr)
		   + (newPeer?" (new)":" (update)"));
	
    	// If attribute '_baddr' is available in the beacon's
    	// descriptor, let us use its value as an indication of the
    	// sender's beaconing address. Otherwise let us consider that
    	// the socket address from which we have received this message
    	// is that of the sender (i.e. value of argument sockAddr)

	InetSocketAddress beaconingAddr = sockAddr;
    	if (baddr != null) 
	    beaconingAddr = DataFormatting.toInetSocketAddress(baddr);
	setPeer(sdr, beaconingAddr);

	// If attribute '_gaddr' is available in the beacon's
    	// descriptor, let us use its value as an indication of the
    	// sender's gossiping address. Otherwise let us the sender
    	// uses the same address for beaconing and gossiping.
	
	InetSocketAddress gossipingAddr = beaconingAddr;
    	if (gaddr != null) 
	    gossipingAddr = DataFormatting.toInetSocketAddress(gaddr);
	
	// Let us see if we must (and can) open a dedicated channel to
	// that peer
	if (use_tcp)
	    if (! manager_.hasPeer(sdr))
		openChannel(sdr, gossipingAddr);
	
	// Let us say 'hello' to any *new* peer, unless it said
	// 'hello' to us first (specifically), which means it
	// obviously knows us
	if ((sender_ != null) && newPeer && (rcv == null)) 
	    sender_.sendAnnounce(sdr);

	if ((manager_ != null) && ! use_tcp)
	    if (newPeer)
		manager_.peerDiscovered(sdr, this);
	    else
		manager_.peerConfirmed(sdr);
    }

    // ------------------------------------------------------------
    private void removePeer(String peerId) {

	if (! peers_.containsKey(peerId))
	    return;

	if (use_tcp)
	    closeChannel(peerId);
	
	peers_.remove(peerId);
	
	if (manager_ != null)
	    manager_.peerLost(peerId);
    }

    // ------------------------------------------------------------
    private void openChannel(String peerId,
			     InetSocketAddress sockAddr) {
	
	int cmp = hostId_.compareTo(peerId);
	Logger.log("udp_chan", "localHost (" + hostId_
		   + ") < remoteHost (" + peerId
		   + ") = " + (cmp < 0));
	if (cmp >= 0)
	    return;
	
	Logger.log("udp_chan", "attempting to open a new channel to "
		   + DataFormatting.toString(sockAddr));

	try {
	    InetSocketAddress localAddr =
		new InetSocketAddress(localSocketAddr_.getAddress(), 0);

	    // Channel channel = new TCP_Channel(localAddr, sockAddr, 5000);
	    Channel channel = new TCP_Channel(sockAddr, 5000, manager_);
	    manager_.addChannel(channel);
	    channel.open();
	}
	catch (Exception e) {
	    Logger.log("udp_chan", "warning: failed to open a TCP_Channel to "
		       + DataFormatting.toString(sockAddr));
	}
    }
	
    // ------------------------------------------------------------
    private void closeChannel(String peerId) {

	if (manager_ == null)
	    return;

	Channel channel = manager_.getChannelForPeer(peerId);
	if (channel != null)
	    channel.close();
    }
	
    // ------------------------------------------------------------
    private void processByeAnnounce(String sdr) {

	Logger.log("udp_chan", "< (bye) sdr=" + sdr);

	removePeer(sdr);
    }

    // ------------------------------------------------------------
    private void checkPeers() {

	if (latency_ == 0)
	    return;
	
	// Logger.log("udp_chan", "checking peers");
	
    	long now = System.currentTimeMillis();
	
    	// Let us check if the local host's registry is up-to-date
    	// (removing peers we haven't heard about for a long time)
    	synchronized(peers_) {
	    // FG: synchro ci-dessus a revoir
    	    Set<String> lostPeers = new HashSet<String>();
    	    for (String id: peers_.keySet()) {
		Peer peer = peers_.get(id);
    		long elapsed = (now - peer.lastRcvd) / 1000;
		// Logger.log("udp_chan", " peer=" + id
		// 	   + ": " + elapsed + " sec");
    		if (elapsed > latency_) 
    		    lostPeers.add(id);
    	    }

	    if (! lostPeers.isEmpty())
		Logger.log("udp_chan", "discarding peers=["
			   + DataFormatting.toString(lostPeers, ",") + "]");
	    
    	    for (String id: lostPeers)
    		removePeer(id);
    	}
    }

    // ------------------------------------------------------------
    private void processData(byte[] buffer)
	throws Exception {

	Message msg = in_bufferizer_.fromBuffer(buffer);
	Descriptor desc = msg.getDescriptor();
	
	String sdr = desc.getAttribute("_sdr");
	// Logger.log("udp_chan", "< ["
	//    	   + buffer.length + "] (sdr=" + sdr 
	//    	   + ") saddr=" + lastRcvAddr_);
	
	// Let us check the sender is identified
	if (sdr == null) {
	    Logger.log("udp_chan", "warning: sdr unspecified");
	    return;
	}
	
	// Let us make sure we do not process our own messages
	if (sdr.equals(hostId_))
	    return;
	
	if (manager_ == null) {
	    Logger.log("udp_chan", "warning: no message processor");
	    return;
	}
	
	String btype = desc.getAttribute("_btype");
	if (btype != null) {
	    if (btype.equals("hello")) {
		String rcv  = desc.getAttribute("_rcv");
		String baddr = desc.getAttribute("_baddr");
		String gaddr  = desc.getAttribute("_gaddr");
		processHelloAnnounce(sdr, rcv, baddr, gaddr, msg, lastRcvAddr_);
	    }
	    else if (btype.equals("bye")) {
		processByeAnnounce(sdr);
	    }
	}
	
	Peer peer = peers_.get(sdr);
	if (peer != null)
	    peer.lastRcvd = System.currentTimeMillis();
	
	if ((btype == null) || (passthrough)) {
	    manager_.process(msg);
	}
    }

    // ------------------------------------------------------------
    public void run() {

	isOpen_ = true;

	if (sender_ != null)
	    sender_.resume();
	
	try {
	    while (true) {
		byte[] buffer = receiveBytes();
		processData(buffer);
	    }
	}
	catch(Exception e) {
	    if (isOpen_)
		Logger.log("udp_chan", "rcv_failure");
	}

	if (isOpen_)
	    close();
	
	isOpen_ = false;
	Logger.log("udp_chan", "terminate");
    }

    // ------------------------------------------------------------
    private byte[] receiveBytes() throws IOException {
	
	DatagramPacket inPacket =
	    new DatagramPacket(new byte[MAX_BUFFER_SIZE],
			       MAX_BUFFER_SIZE);
	
	socket_.receive(inPacket);
	    
	// lastRcvTime_ = System.currentTimeMillis();
	
	int nbBytes = inPacket.getLength();
	int offset = inPacket.getOffset();
	lastRcvAddr_ = (InetSocketAddress)inPacket.getSocketAddress();

	// Logger.log("udp_chan", "< ["
	// 	   + nbBytes + "] " 
	// 	   + lastRcvAddr_);
	
	byte[] data = inPacket.getData();
	
	// Checking magic cookie in payload
	if (data[0] != COOKIE) {
	    System.out.println("udp_chan.receiveBytes(): invalid cookie rcvd from "
			       + inPacket.getSocketAddress());
	    throw new IOException("Invalid cookie");
	}
	
	byte[] result = new byte[nbBytes-1];
	DataBufferizer.copyBuffer(data, offset+1, result, 0, nbBytes-1);
	
	return result;
    }

    // ------------------------------------------------------------
    private void sendBytes(InetSocketAddress addr, byte[] buffer)
	throws IOException {
	
	try {
	    DatagramPacket outPacket =
		new DatagramPacket(new byte[buffer.length+1],
				   buffer.length+1);
	    outPacket.setSocketAddress(addr);
	    byte[] data = outPacket.getData();
	    
	    // Inserting magic cookie in buffer
	    data[0] = COOKIE;
	    
	    // Copying data in buffer after the magic cookie
	    DataBufferizer.copyBuffer(buffer, 0, data, 1, buffer.length);
	    
	    // Sending packet
	    outPacket.setLength(buffer.length+1);
	    socket_.send(outPacket);
	    
	    // lastSndTime_ = System.currentTimeMillis();
	    
	    // Logger.log("udp_chan", "> ["
	    // 	       + outPacket.getLength() + "] " 
	    // 	       + outPacket.getSocketAddress());
	    
	} catch (IOException e) {
	    System.out.println("UDP_Channel: could not send packet");
	    throw e;
	}
    }
	
    // ------------------------------------------------------------
    public void open() {

	if (isOpen_)
	    return;
	
	Logger.log("udp_chan", "open "
		   + this);
	
	start();
    }
    
    // ------------------------------------------------------------
    public void close() {

	if (! isOpen_)
	    return;
	
	Logger.log("udp_chan", "close " + this);

	if (sender_ != null)
	    sender_.suspend();
	isOpen_ = false;
	
	try {
	    this.interrupt();
	    socket_.close();
	}
	catch (Exception e) {}
	
	if (use_tcp) 
	    tcp_server_.close();
	
	if (manager_ != null) {
	    for (String peerId: peers_.keySet())
		manager_.peerLost(peerId);
	    manager_.channelClosed(this);
	}

	peers_.clear();
    }

    // ------------------------------------------------------------
    public boolean isOpen() {

	return isOpen_;
    }

    // ------------------------------------------------------------
    public void write(Message msg) {

	if (! isOpen_ ) {
	    Logger.log("udp_chan", "warning: channel closed");
	    return;
	}

	Descriptor desc = msg.getDescriptor();

	// Looking for a "_brcv" (blind receiver) attribute, and if
	// not present for a "_rcv" attribute.
	//
	// The "blind receiver" mechanism is used by a Hub to relay a
	// broadcast message (such as a "hello" or "bye" beacon) to a
	// specific destination
	String rcv = desc.getAttribute("_brcv");
	if (rcv != null)
	    desc.removeAttribute("_brcv");
	else
	    rcv = desc.getAttribute("_rcv");

	synchronized(this) {
	    try {
		
		// If this message is a beacon, then let us remove
		// obsolete peers (if any) before sending this beacon
		if (rcv == null)
		    checkPeers();
		
		if ((rcv == null) && (remoteSocketAddr_ == null)) {
		    Logger.log("udp_chan", "warning: rcv=null and no default target address");
		    return;
		}

		Peer peer = null;
		InetSocketAddress targetAddr;
		if (rcv != null) {
		    peer = peers_.get(rcv);
		    if (peer == null) {
			Logger.log("udp_chan", "warning: unknown peer="
				   + rcv);
			return;
		    }
		    targetAddr = peer.address;
		}
		else
		    targetAddr = remoteSocketAddr_;

		byte[] buffer = out_bufferizer_.toBuffer(msg);

		// Logger.log("udp_chan", "> [" + buffer.length + "] (rcv="
		//    	   + rcv + ") daddr=" + DataFormatting.toString(targetAddr));

		sendBytes(targetAddr, buffer);
		if (peer != null)
		    peer.lastSent = System.currentTimeMillis();
	    }
	    catch (Exception e) {
		Logger.log("udp_chan", "warning: tx failed (rcv="
			   + rcv + ")");
		// Logger.log("udp_chan", e.toString());
	    }
	}
    }

    // ------------------------------------------------------------
    public String toString() {

	return "UDP_Channel(local="
	    + DataFormatting.toString(localSocketAddr_)
	    + ",remote="
	    + DataFormatting.toString(remoteSocketAddr_)
	    + (intfAddr_==null?"":",iface=" + DataFormatting.toString(intfAddr_))
	    + ",period=" + period_
	    + (tcp_server_!=null?",use_tcp":"")
	    + ")";
    }
}
