/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.util;

import java.io.InputStream;
import java.io.FileInputStream;
import java.io.PrintStream;
import java.io.File;
import java.io.IOException;

import casa.dodwan.util.BufferAccessException;

/**
 * A <code>Bufferizer&lt;E&gt;</code> implements a strategy for storing an
 * object of type <code>E</code> in a buffer (see method <code>toBuffer()</code>
 * ), and for extracting an object of type <code>E</code> from a buffer (see
 * method <code>fromBuffer()</code>).
 * 
 * @author $Author: F. Guidec$
 * @version $Revision: 1.0$
 */

public abstract class Bufferizer<E> {
    
    // ------------------------------------------------------------
    /**
     * Extracts an object of type <code>E</code> from the specified buffer,
     * starting at the specified offset in this buffer
     * 
     * @param buffer
     *            the buffer an object must be extracted from
     * @param offset
     *            index of the first byte to be read from this buffer
     * @param length
     *            the number of bytes to be read from this buffer
     * 
     * @return the object that was stored in the buffer
     * 
     * @throws BufferAccessException
     *             when no object could be extracted from the buffer, starting
     *             at the specified <code>offset</code>
     */
    public abstract E fromBuffer(byte[] buffer, int offset, int length)
	throws BufferAccessException;
    

    // ------------------------------------------------------------
    public E fromBuffer(byte[] buffer)
	throws BufferAccessException {

	return fromBuffer(buffer, 0, buffer.length);
    }

    // ------------------------------------------------------------
    public E fromFile(File file) throws BufferAccessException,
	IOException {
	
	InputStream in = new FileInputStream(file);
	int fileLength = (int) file.length();
	byte[] buffer = new byte[fileLength];
	int nread = in.read(buffer, 0, fileLength);
	in.close();
	
	E result = fromBuffer(buffer, 0, fileLength);
	return result;
    }

    // ------------------------------------------------------------
    public E fromFile(String fileName) throws BufferAccessException,
	IOException {
	
	File file = new File(fileName);
	
	return fromFile(file);
    }
    
    // ------------------------------------------------------------
    /**
     * Stores an object of type <code>E</code> at the specified offset position
     * in the specified buffer.
     * 
     * @param value
     *            a reference to the object that must be stored in the buffer
     * @param buffer
     *            the buffer this object must be stored in
     * @param offset
     *            index of the first byte where <code>value</code> can be stored
     *            in <code>buffer/code>
     * 
     * @return the number of bytes used in the buffer, starting at position
     *         <code>offset</code>
     * 
     * @throws BufferAccessException
     *             when there is not enough space in the buffer to hold ths
     *             object
     */
    public abstract int toBuffer(E value, byte[] buffer, int offset)
	throws BufferAccessException;
    
    
    // ------------------------------------------------------------
    public abstract byte[] toBuffer(E value)
	throws BufferAccessException;

    // ------------------------------------------------------------
    public void toFile(E obj, File file)
	throws BufferAccessException, IOException {
	
	PrintStream out = new PrintStream(file);
	
	byte[] buf = toBuffer(obj);
	out.write(buf);
	out.flush();
	out.close();	
    }
    
    // ------------------------------------------------------------
    public void toFile(E obj, String fileName)
	throws BufferAccessException, IOException {
	
	File file = new File(fileName);

	toFile(obj, file);
    }

}
