/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.util;

import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;
import java.util.BitSet;
import java.util.Collection;

public class Catalog {
    
    private Map<String,BitSet> map_ =
	new HashMap<String,BitSet>();

    // ------------------------------------------------------------
    public int size() {
	return map_.size();
    }

    // ------------------------------------------------------------
    public void clear() {
	map_.clear();
    }

    // ------------------------------------------------------------
    public boolean isEmpty() {
	return map_.isEmpty();
    }

    // ------------------------------------------------------------
    public boolean isEmpty(String key) {
	return map_.get(key).isEmpty();
    }

    // ------------------------------------------------------------
    public Set<String> keySet() {
	return map_.keySet();
    }

    // ------------------------------------------------------------
    public boolean contains(String key) {
	return map_.containsKey(key);
    }

    // ------------------------------------------------------------
    public void put(String key) {
	BitSet fragments = new BitSet();
	map_.put(key, fragments);
    }
    
    // ------------------------------------------------------------
    public void put(String key, BitSet fragments) {
	map_.put(key, fragments);
    }
    
    // ------------------------------------------------------------
    public void set(String key, int fragId) {
	if (map_.containsKey(key))
	    map_.get(key).set(fragId);
	else {
	    BitSet fragments = new BitSet();
	    fragments.set(fragId);
	    map_.put(key, fragments);
	}
    }
    
    // ------------------------------------------------------------
    public void clear(String key, int fragId) {
	if (map_.containsKey(key))
	    map_.get(key).clear(fragId);
	else {
	    BitSet fragments = new BitSet();
	    map_.put(key, fragments);
	}
    }
    
    // ------------------------------------------------------------
    public BitSet getFragments(String key) {
	return map_.get(key);
    }
    
    // ------------------------------------------------------------
    public Object clone() {
	
	Catalog result = new Catalog();
	for (String key: map_.keySet()) {
	    BitSet bitset = map_.get(key);
	    result.put(key, (BitSet)bitset.clone());
	}
	return result;
    }

    // ------------------------------------------------------------
    public void add(Catalog other) {

	if (other == null)
	    return;

	for (String key: other.map_.keySet()) {
	    if (! map_.containsKey(key)) 
		map_.put(key, other.map_.get(key));
	    else
		map_.get(key).or(other.map_.get(key));
	}
    }

    // ------------------------------------------------------------
    public void remove(String key) {
	map_.remove(key);
    }
    
    // ------------------------------------------------------------
    public void remove(Collection<String> other) {
	for (String key: other)
	    map_.remove(key);
    }
    
    // ------------------------------------------------------------
    public void remove(Catalog other) {

	if (other == null)
	    return;

	for (String key: other.map_.keySet()) {
	    if (map_.containsKey(key)) {
		map_.get(key).andNot(other.map_.get(key));
		if (map_.get(key).isEmpty())
		    map_.remove(key);
	    }
	}
    }

    // ------------------------------------------------------------
    public void retainAll(Collection<String> other) {

	if (other == null)
	    return;

	Set<String> toRemove = new HashSet<String>();
	for (String key: map_.keySet()) 
	    if (! other.contains(key)) 
		toRemove.add(key);
	this.remove(toRemove);
    }

    // ------------------------------------------------------------
    public String toString() {

	String result = "";
	
	boolean firstItem = true;
	for (String key: map_.keySet()) {
	    BitSet bitset = map_.get(key);
	    if (! firstItem) 
		result += ",";
	    else
		firstItem = false;
	    result += key + "[";
	    for (int i=0; i < bitset.length(); i++)
		if (bitset.get(i))
		    result += "1";
		else
		    result += "0";
	    result += "]";
	}
	return result;
    }

    // ------------------------------------------------------------
    public static void main(String[] args) {

	Catalog c1 = casa.dodwan.util.DataFormatting.splitToCatalog("b1[01010101],b2[1010101]", ",");
	Catalog c2 = casa.dodwan.util.DataFormatting.splitToCatalog("b3[1111111],b2[1111]", ",");

	System.out.println("c1\t\t\t= " + c1);
	System.out.println("c2\t\t\t= " + c2);

	Catalog c3 = (Catalog)c1.clone();
	System.out.println("c3\t\t\t= " + c3);

	c3.add(c2);
	System.out.println("c3.add(c2)\t\t= " + c3);
	
	c3.remove(c1);
	System.out.println("c3.remove(c1)\t\t= " + c3);

	c3.retainAll(c1.keySet());
	System.out.println("c3.retainAll(c1)\t= " + c3);
    }
}
