/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/




package casa.dodwan.util;
 
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Properties;


/**
 * A <code>ConfigurationProperties</code> object holds Java properties
 * dedicated to an application service. It is initialized specifically for
 * this service and searched for typed values.  Properties are initialized from
 * a default configuration file, overriden by definitions found in a user
 * configuration file and then by system properties (i.e. defined on the
 * command line when launching the JVM).  Properties can be written back to the
 * user configuration file.  <p> The property keys in the configuration files
 * are assumed to start with a fixed string specific to the service (the name
 * of the service).  In system properties, properties whose key does not start
 * with this prefix are ignored.
 *
 * <p> Property values in configuration files may contain references to system
 * properties.  In this case, variable substitution is applied. For example you may
 * find in a configuration file a definition of the form <code>foo =
 * ${user.home}/bar</code>. Note that substitutions are performed only once.
 * So references embedding other references are not allowed. 
 *
 * <p>The default configuration file may contain a specific property definition
 * with the key </code>immutable</code>.  The value of this property is a
 * comma-separated list of property keys. The definition of the properties
 * whose key appears in this list cannot be overriden in the user configuration
 * file or through system properties. 
 */
public class ConfigurationProperties {

    private String userConfFileName_ = null;
 
    // The properties read from the default file
    // Filtered according to the service and and with refs substituted
    private Properties defaultProps_;
    
    // The properties made available (include defaultProps_ as defaults)
    // Filtered according to the service and and with refs substituted
    private Properties confProps_;   

    // The properties as they are read from the user conf file
    private Properties userProps_ = null;       

    // The set of the keys of the modified configuration properties
    // i.e. those whose setter has been called
    private HashSet<String> modifiedKeys_ ;
    


    //------------------------------------------------------------------------------------
    /**
     * Creates a new <code>ConfigurationProperties</code> object. Initial values are 
     * @param ServiceAppPrefix name of the application service. 
              This name is a mandatory prefix for keys of the properties to be read.
     * @param defaultConfFileName name of the default configuration file (normally, a read-only file).
     * @param userConfFileName name of the user configuration file (normally a read-write file). 
     */
    public ConfigurationProperties(String ServiceAppPrefix, String defaultConfFileName, String userConfFileName) {
    	System.out.println("Create ConfigurationProperties("+ServiceAppPrefix+", "+defaultConfFileName+", "+userConfFileName);
    	
	userConfFileName_ = userConfFileName;

	defaultProps_ = new Properties();
	userProps_ = new Properties();
	modifiedKeys_ = new HashSet<String>();

	// Read the properties in the default configuration file if it exists
	if (defaultConfFileName != null) {
	    File f = new File(defaultConfFileName);
	    if (f.exists()) {
		try {
		    System.out.println("ConfigurationProperties() : Reading default configuration file " + defaultConfFileName);
		    FileInputStream fs = new FileInputStream(f);
		    defaultProps_.load(fs);
		    fs.close();
		}
		catch (Exception e) {
		    System.out.println("ConfigurationProperties() :  Cannot read default configuration file " + defaultConfFileName);
		}
	    }
	}
	replaceVariables(defaultProps_);

	confProps_ = new Properties(defaultProps_);

	// Build the set of immutable property keys from the value of the "immutable" property in the .default
	HashSet<String> immutables = new HashSet<String>();
	String immutableProp = confProps_.getProperty("immutable");
	if (immutableProp != null) {
	    String[] simm = immutableProp.split(",");
	    for (int i=0; i < simm.length; i++) {
		immutables.add(simm[i].trim());
	    }
	}
	
        
	// Complete confProps_ (and so override) with properties from the user configuration file if it exists
	if (userConfFileName != null) {
	    File f = new File(userConfFileName);
	    if (f.exists()) {
		try {
		    // Read the user conf properties
		    System.out.println("ConfigurationProperties() : Reading user configuration file " + userConfFileName);
		    FileInputStream fs = new FileInputStream(f);
		    userProps_.load(fs);
		    fs.close();
		    
		    // Consider only the right properties
		    complete(confProps_, userProps_, ServiceAppPrefix, immutables);
		}
		catch (Exception e) {
		    System.out.println("ConfigurationProperties() :  Cannot read user configuration file " + userConfFileName);
		}
	    }
	}
        // Complete confProps_ (and possibly override) with the right system properties
	complete(confProps_, System.getProperties(), ServiceAppPrefix, immutables);
    }





    // ------------------------------------------------------------------------------------------
    /*
     * Completes (and possibly overrides) a set of properties with another one
     * The properties added are those that start with the right prefix and are not a priori excluded
     * Variable substitution is performed on the added properties.
     * @param org The properties to be completed
     * @param other The properties from which some are added to org
     * @param prefix The prefix of the properties to consider in other
     * @param excluded The set of the keys of the properties from other not to add
     */  

    private void complete(Properties org, Properties other, String prefix, HashSet<String> excluded) {
	for (Enumeration e = other.keys(); e.hasMoreElements(); ) {
	    String key = (String) e.nextElement();
	    if ( key.startsWith(prefix) && (! excluded.contains(key)) ) {
		String value = replaceVariables(other.getProperty(key));
		org.setProperty(key,value);
	    }
	}
    }



    // ------------------------------------------------------------------------------------------
    /*
     * Replaces in all property values the references to variables with the values taken from system properties
     * References are of the form ${[^}]*}. The substitutions are performed only once.
     * So references embedding other references are not allowed.
     * @param str The string to examine
     * @return The string with the replacement done 
     */
    private void replaceVariables(Properties props) {
	for (Enumeration e = props.keys(); e.hasMoreElements(); ) {
	    String key = (String) e.nextElement();
	    String value = replaceVariables(props.getProperty(key));
	    props.setProperty(key, value);
	}
    }



    // ------------------------------------------------------------------------------------------
    /*
     * Replace in a string the references to variables with the values taken from system properties
     * (If a variable is not in the system properties, the reference isn't replaced.)
     * References are of the form ${[^}]*}. The substitutions are performed only once.
     * So references embedding other references are not allowed.
     * 
     * @param str The string to examine
     * @return The string with the replacement done 
     */  
    private String replaceVariables(String str) {
	
	String buf = str;

	int start = buf.indexOf("${");
	while (start != -1) { 
	    int end = buf.indexOf("}", start+2);
	    if (end == -1) break;
	    if (start == (end -2)) {
		// ref vide
		start = buf.indexOf("${", end);
	    }
	    else {
		String ref = buf.substring(start+2 , end);
		String value  = System.getProperty(ref);
		if (value != null) {
		    buf =   buf.substring(0, start) 
			+ value 
			+ buf.substring(end+1, buf.length());
		    start = buf.indexOf("${", start + value.length());
		}
		else {
		    start = buf.indexOf("${", end);
		}
	    }
	}
	return buf;
    }


    //------------------------------------------------------------------------------------
    /**
     * Saves all the configuration properties in the user configuration file 
     * The previous content of this file is lost. 
     * Does nothing if the name of the configuration file, as passed in the constructor, is <tt>null</tt>.
     * @throws IOException if an error occurs when writing in the file.
     */
    public void saveConfiguration() throws  IOException {
	if (userConfFileName_ != null) {
	    System.out.println("ConfigurationProperties()saveConfiguration(): Writing file " + userConfFileName_);
	    FileOutputStream fs = new FileOutputStream(userConfFileName_);
	    confProps_.store(fs, null);
	}
    }
	


    //------------------------------------------------------------------------------------
    /** 
     * Augment the user configuration file with properties that
     * have been modified since the last call to this method.
     * Does nothing if the name of the configuration file, as passed in the constructor, is <tt>null</tt>.
     * @throws IOException if an error occurs when writing in the file.
     */
    public void saveUserConfiguration() throws  IOException {
	if (userConfFileName_ != null && ( ! modifiedKeys_.isEmpty())) {
	    System.out.println("ConfigurationProperties().saveUserConfiguration(): Writing file " + userConfFileName_);
	    FileOutputStream fs = new FileOutputStream(userConfFileName_);
	    
	    for (String key: modifiedKeys_) {
		userProps_.setProperty(key, confProps_.getProperty(key));
	    }
	    
	    userProps_.store(fs, null);

	    modifiedKeys_.clear();
	}
    }


 
    //----- Getters ----------------------------------------------------------------------

    //------------------------------------------------------------------------------------
    /** 
     * Looks for a property of type <code>String</code> and returns its value.
     * @param key the key of the property to look for.
     * @param defaultValue the value to return if the property is not found.
     */
    public String getString(String key, String defaultValue) {
	String s = confProps_.getProperty(key);
        String ret = defaultValue;
	if (s != null) {
	    ret = s;
	}
	return ret;
    }

    //------------------------------------------------------------------------------------
    /** 
     * Looks for a property whose value represents an integer, and returns its value.
     * If no property is found or if an error occurs when parsing the found value, 
     * a default value is returned.
     * @param key the key of the property to look for. 
     * @param defaultValue the value to return if the property is not found.
     * @return the value of the property as an <code>int</code>. 
     */
    public int getInteger(String key, int defaultValue) {
	String s = confProps_.getProperty(key);
        int ret = defaultValue;
	if (s != null) {
	    try {
		ret = Integer.parseInt(s);
	    }
	    catch (NumberFormatException e) {
	    }
	}
	return ret;
    }

   //------------------------------------------------------------------------------------
    /** 
     * Looks for a property whose value represents an quantified integer
     * (e.g. 4kB), and returns its value.  Recognised quantifiers are
     * <code>B</code> for bytes, <code>k</code> or <code>kB</code> for
     * kilo-bytes, <code>M</code> or <code>MB</code> for mega-bytes.  
     * The quantifier is optional. If no
     * property is found or if an error occurs when parsing the found value, the specified 
     * default value is returned.
     * @param key the key of the property to look for. 
     * @param defaultValue the value to return if the property is not found.
     * @return the value of the property as an <code>int</code>.
     */
    public int getQuantifiedInteger(String key, int defaultValue) {
	String s = confProps_.getProperty(key);
        int ret = defaultValue;

	if (s == null) 
	    return ret;

	s = s.toLowerCase();
	int len = s.length();
	int quant = 1;

//  	System.out.println("getQuantifier(" + s + ")");
//  	System.out.println("  low=" + s + ", len=" + len);

	try {
	    if (s.endsWith("mb")) 
		return 1048576 * Integer.parseInt(s.substring(0, len-2));
	    
	    if (s.endsWith("m")) 
		return 1048576 * Integer.parseInt(s.substring(0, len-1));
	    

	    if (s.endsWith("kb")) 
		return 1024 * Integer.parseInt(s.substring(0, len-2));
	    
	    if (s.endsWith("k")) 
		return 1024 * Integer.parseInt(s.substring(0, len-1));

	    if (s.endsWith("b")) 
		return Integer.parseInt(s.substring(0, len-1));

	    return Integer.parseInt(s);
	}
	catch (NumberFormatException e) {}
	
	return ret;
    }

    //------------------------------------------------------------------------------------
    /** 
     * Looks for a property whose value represents a long integer, and returns its value.
     * If no property is found or if an error occurs when parsing the found value, 
     * a default value is returned.
     * @param key the key of the property to look for. 
     * @param defaultValue the value to return if the property is not found.
     * @return the value of the property as a <code>long</code>. 
     */
   public long getLong(String key, long defaultValue) {
	String s = confProps_.getProperty(key);
        long ret = defaultValue;
	if (s != null) {
	    try {
		ret = Long.parseLong(s);
	    }
	    catch (NumberFormatException e) {
	    }
	}
	return ret;
    }

    //------------------------------------------------------------------------------------
   /** 
     * Looks for a property whose value represents a boolean, and returns its value.
     * If no property is found or if an error occurs when parsing the found value, 
     * a default value is returned.
     * @param key the key of the property to look for. 
     * @param defaultValue the value to return if the property is not found.
     * @return the value of the property as a <code>long</code>. 
     */
    public boolean getBoolean(String key, boolean defaultValue) {
	String s = confProps_.getProperty(key);
        boolean ret = defaultValue;
	if (s != null) {
	    try {
		ret = Boolean.parseBoolean(s);
	    }
	    catch (NumberFormatException e) {
	    }
	}
	return ret;
    }

    //------------------------------------------------------------------------------------
    /** 
     * Looks for a property whose value represents a real number, and returns its value.
     * If no property is found or if an error occurs when parsing the found value, 
     * a default value is returned.
     * @param key the key of the property to look for. 
     * @param defaultValue the value to return if the property is not found.
     * @return the value of the property as a <code>double</code>. 
     */
   public double getDouble(String key, double defaultValue) {
	String s = confProps_.getProperty(key);
        double ret = defaultValue;
	if (s != null) {
	    try {
		ret = Double.parseDouble(s);
	    }
	    catch (NumberFormatException e) {
	    }
	}
	return ret;
    }

 

    //----- Setters ----------------------------------------------------------------------

    //------------------------------------------------------------------------------------
    /**
     * Sets the value of a property.
     * @param key the key of the property to set.
     * @param value the new value of the property.
     */
   public void setString(String key, String value) {
	confProps_.setProperty(key, value);
	modifiedKeys_.add(key);
    }

    //------------------------------------------------------------------------------------
     /**
     * Sets the value of a property from an integer.
     * @param key the key of the property to set.
     * @param value the new value of the property.
     */
   public void setInteger(String key, int value) {
	confProps_.setProperty(key, Integer.toString(value));
	modifiedKeys_.add(key);
     }

    //------------------------------------------------------------------------------------
     /**
     * Sets the value of a property from an long integer.
     * @param key the key of the property to set.
     * @param value the new value of the property.
     */
    public void setLong(String key, long value) {
	confProps_.setProperty(key, Long.toString(value));
 	modifiedKeys_.add(key);
    }

    //------------------------------------------------------------------------------------
      /**
     * Sets the value of a property from a boolean.
     * @param key the key of the property to set.
     * @param value the new value of the property.
     */
   public void setBoolean(String key, boolean value) {
	confProps_.setProperty(key, Boolean.toString(value));
	modifiedKeys_.add(key);
     }

    //------------------------------------------------------------------------------------
    /**
     * Sets the value of a property from a real number.
     * @param key the key of the property to set.
     * @param value the new value of the property.
     */
     public void setDouble(String key, double value) {
	confProps_.setProperty(key, Double.toString(value));
	modifiedKeys_.add(key);
     }





    //------------------------------------------------------------------------------------
    /**
     * @undocumented
     */
     public void dump() {
	System.out.println("############# Default Configuration properties :");
	list(defaultProps_);

	System.out.println("############# User Configuration properties :");
	list(userProps_);

	System.out.println("############# Configuration properties :");
	list(confProps_);
    }
     

    //------------------------------------------------------------------------------------
    /**
     * @undocumented
     */
   private void list(Properties props) {
	for (Enumeration en = props.keys(); en.hasMoreElements(); ) {
	    String key = (String) en.nextElement();
	    String value = props.getProperty(key);
	    System.out.println(key + " = " + value);
	}
    }





}


 
