/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.util;

/**
 * A <code>Filter&lt;E&gt;</code> is a <code>Sink&lt;E&gt;</code> that can feed
 * another <code>Sink&lt;E&gt;</code>, while filtering out objects that do match
 * a given selection criterion.
 * 
 * <p>
 * 
 * A <code>Filter</code> is associated with an outgoing <code>Sink</code>, and a
 * <code>Selector</code>. Any object written in the <code>Filter</code> is
 * matched against the selection criterion defined in the <code>Selector</code>,
 * and only those objects that match this criterion are written to the outgoing
 * <code>Sink</code>.
 * 
 * @author $Author: F. Guidec$
 * @version $Revision: 1.0$
 */
public class Filter<E> implements Sink<E>, SinkListener<E> {

	private Selector<E> selector_;

	private Sink<E> sink_;

	public Verbosity verbosity = new Verbosity();

	/**
	 * Creates a new <code>Filter</code>, and associates this filter with an
	 * outgoing <code>Sink</code> and with a <code>Selector</code>
	 * 
	 * @param sink
	 *            the outgoing <code>Sink</code> this <code>Filter</code> is
	 *            associated with
	 * @param selector
	 *            the <code>Selector</code> this <code>Filter</code> should use
	 *            to filter out objects that must not be written to the outgoing
	 *            <code>Sink</code>
	 */
	public Filter(Sink<E> sink, Selector<E> selector) {
		setSink(sink);
		setSelector(selector);
	}

	/**
	 * Returns a reference to the <code>Selector</code> this <code>Filter</code>
	 * is associated with
	 * 
	 * @return the <code>Selector</code> this <code>Filter</code> is associated
	 *         with
	 */
	public Selector<E> getSelector() {
		return selector_;
	}

	/**
	 * Returns a reference to the outgoing <code>Sink</code> this
	 * <code>Filter</code> is associated with
	 * 
	 * @return the outgoing <code>Sink</code> this <code>Filter</code> is
	 *         associated with
	 */
	public Sink<E> getSink() {
		return sink_;
	}

	/**
	 * Changes the <code>Selector</code> this <code>Filter</code> should use to
	 * filter out undesired objects
	 * 
	 * @param selector
	 *            the <code>Selector</code> to be used for filtering out
	 *            undesired objects
	 */
	public void setSelector(Selector<E> selector) {
		selector_ = selector;
	}

	/**
	 * Changes the <code>Sink</code> this <code>Filter</code> should write
	 * outgoing objects in
	 * 
	 * @param sink
	 *            the <code>Sink</code> this <code>Filter</code> should write
	 *            outgoing objects in
	 */
	public void setSink(Sink<E> sink) {
		sink_ = sink;
	}

	/**
	 * Writes an object in this <code>Filter</code>. This object is then matched
	 * against the selection criterion defined in the <code>Selector</code> this
	 * filter is associated with, and if the selection criterion is matched,
	 * then the object is written to the outgoing <code>Sink</code>
	 * 
	 * @param value
	 *            the object that should be written in the outgoing
	 *            <code>Sink</code>, provided it matches the
	 *            <code>Selector</code> of this filter
	 */
	public void write(E value) throws Exception {

		boolean matched = ((selector_ == null) || selector_.isMatchedBy(value));

		if (verbosity.isEnabled()) {
			System.out.println("Filter.write()" + value + ")");
			System.out.println("  Selector is matched: " + matched);
		}

		if (matched)
			sink_.write(value);

	}

}
