/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.util;

import java.io.File;
import java.util.regex.Pattern;

import casa.dodwan.run.Dodwan;
import casa.dodwan.docware.Descriptor;
import casa.dodwan.util.Logger;

public class LogManagement 
    implements Processor<Descriptor> {

    private File path_;
    private Dodwan dodwan_;

    // ------------------------------------------------------------
    public LogManagement(Dodwan dodwan, File path) {

	dodwan_ = dodwan;
	path_ = path;

	if (!path_.exists())
            path_.mkdirs();

	// Create a pattern for logging requests and replies, and set
	// a subscription accordingly
	Descriptor logPattern = new Descriptor();
	logPattern.setAttribute("_stype", "logging/.*");
	logPattern.setDestination(SystemEnvironment.host + "|ANY");

	dodwan_.pubSubService.addSubscription("log_mgt", logPattern, this);
    }

    // ------------------------------------------------------------
    public void sendRequest(String host, String fpattern, long lifetime) {


	Descriptor desc = new Descriptor();
	desc.setSource(SystemEnvironment.host);
	desc.setDestination(host);
	desc.setAttribute("_stype", "logging/request");
	desc.setAttribute("_fpattern", fpattern);
	desc.setLifetime(lifetime);

	try {
	    dodwan_.pubSubService.publish(desc, new byte[1]);
	}
	catch (Exception e) {
	    System.out.println(e);
	}

	Logger.log("log_mgt", "> dst=" + host
		   + ",pattern=" + fpattern);
    }

    // ------------------------------------------------------------
    public void process(Descriptor desc) {

	if (desc == null) 
	    return;
	
	// System.out.println("LogManagement: process()");

	// Check the type in this descriptor
	String type = desc.getAttribute("_stype");
	if ((type == null) || (! type.startsWith("logging/")))
	    return;

	if (type.equals("logging/reply")) {
	    processReply(desc);
	    return;
	}

	if (type.equals("logging/request")) {
	    processRequest(desc);
	    return;
	}
    }

    // ------------------------------------------------------------
    public void processReply(Descriptor desc) {

	// System.out.println("LogManagement: processReply()");

	String key = desc.getDocumentId();
	String fname = desc.getAttribute("_fname");
	String src = desc.getSource();

	File file = new File(path_, fname);
	try {
	    // if (! file.exists())
		dodwan_.pubSubService.getAsFile(key, file);
	    // else
	    // 	System.out.println("  LogManagement: dup");
	}
	catch (Exception e) {
	    System.out.println(e);
	}

	Logger.log("log_mgt", "< src=" + src
		   + ",file=" + fname);
    }

    // ------------------------------------------------------------
    public void processRequest(Descriptor desc) {

	// System.out.println("LogManagement: processRequest()");

	String src = desc.getSource();
	if (src == null)
	    return;

	if (src.equals(SystemEnvironment.host))
	    return;

	String fpattern = desc.getAttribute("_fpattern");
	java.util.Date deadline = desc.getDeadline();

	Logger.log("log_mgt", "< src=" + src
		   + ",pattern=" + fpattern);

	// System.out.println("LogManagement:   src=" + src
	// 		   + ",fpattern=" + fpattern
	// 		   + ",deadline=" + deadline);

	File[] files = path_.listFiles();
	// System.out.println("LogManagement:   Considering files in " + path_);

	for (File file: files) {
	    String fname = file.getName();

	    // System.out.println("LogManagement:   Considering file " + fname);
	    if (Pattern.matches(fpattern, fname)) {
		
		try {
		    Logger.log("log_mgt", "> dst=" + src
			       + ",file=" + fname);
		    
		    // System.out.println("LogManagement:     Publishing file " + file);
		    Descriptor desc2 = new Descriptor();
		    // desc2.setDocumentId(MD5sum.digest(fname));
		    desc2.setSource(SystemEnvironment.host);
		    desc2.setDestination(src);
		    desc2.setAttribute("_stype", "logging/reply");
		    desc2.setAttribute("_fname", fname);
		    desc2.setDeadline(deadline);
		    
		    dodwan_.pubSubService.publish(desc2, file);
		}
		catch (Exception e) {
		    System.out.println(e);
		}
		
	    }
	}
    }

}
