/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/




package casa.dodwan.util;

import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.text.SimpleDateFormat;
import java.util.Locale;
import java.net.InetSocketAddress;

/**
 * Provides facilities for logging event messages on file.
 * Event messages are of the form <em>date</em> <em>prefix</em>  <em>tag</em> <em>message</em>.
 * The prefix is common to all messages in a file. 
 * Dates are written with the format <em>yyyy-MM-dd HH:mm:ss.SSS</em>. 
 */
public class Logger {

    private static PrintStream ps_ = null;
    private static Logger logger_ = null;
    private static String name_ = null;
    
    private static final SimpleDateFormat dateFormat = 
	new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSS", Locale.ENGLISH);
    
    // ------------------------------------------------------------
    /**
     * Clear the specified logging directory, i.e. delete any file it
     * may contain.
     * @param dir the directory whose files will be deleted
     */
    public static void clear(File dir) {

	if ((dir == null) || ! dir.isDirectory())
	    return;
	
	if (ps_ != null)
	    disable();

	try {
	    File[] files = dir.listFiles();
	    for (File file: files)
		file.delete();
	}
	catch (Exception e) {}
    }
    
    // ------------------------------------------------------------
    /**
     * Upload all log files found in the specified directory to the
     * specified target (identified by a socket address).
     * @param dir the directory whose files will be deleted
     * @param sockAddr socket address of the target host
     */
    public static void upload(File dir, InetSocketAddress sockAddr) {

	if ((dir == null) || ! dir.isDirectory())
	    return;
	
	try {
	    File[] files = dir.listFiles();
	    for (File file: files)
		FileTransfer.sendFile(file, sockAddr);
	}
	catch (Exception e) {}
    }
    
    // ------------------------------------------------------------
    /**
     * Enable logging, that is prepares the logging system to write log messages. 
     * @param file the file to which log messages will be written
     * @param append if true, then messages will be written to the end of the file rather than the beginning
     * @param name the prefix of each log message (generally used as the source of messages)
     */
    public static void enable(File file, boolean append, String name){
	
	if (ps_ != null)
	    disable();
	
	//  	System.out.println("Logger.enable(" + file + ")");
	
	name_ = name;
	
	if (file == null) {
	    ps_ = System.out;
	    return;
	}
	
	try {
	    
	    File dir = file.getParentFile();
	    if (dir != null)
		dir.mkdirs();
	    
	    ps_ = new PrintStream(new FileOutputStream(file, append));
	}
	catch (Exception e) {
	    System.out.println("Logger(" + file.getAbsolutePath() 
			       + ", " + append
			       + "): failed to open file");
	    System.out.println("  Will log to stdout instead");
	    ps_ = null;
	}
    }
    
    // ------------------------------------------------------------
    /**
     * Disables logging, that is , makes the logging system close the 
     * current log file and forbid logging
     */
    public static void disable() {
	
	//  	System.out.println("Logger.disable()");
	
	if (ps_ != null) 
	    try {
		ps_.close();
	    }
	    catch (Exception e) {}
	
	ps_ = null;
    }
    
    // ------------------------------------------------------------
    /**
     * Disables logging, that is , makes the logging system close the 
     * current log file and forbid logging
     */
    public static boolean isEnabled() {
	
	return (ps_ != null);
    }
    
    // ------------------------------------------------------------
    // private static long lastLog_ = 0;
    
    
    // ------------------------------------------------------------
    /**
     * Log a new event message in the current log file
     * @param tag the tag associated with the message
     * @param msg the event message itself
     */
    public static void log(String tag, String msg) {
	
	if (ps_ == null) 
	    return;
	
	synchronized(ps_) {
	    long now = System.currentTimeMillis();
	    
	    //  	    if (now - lastLog_ > 1000)
	    //  		ps_.println();
	    //  	    lastLog_ = now;
	    
	    ps_.println(dateFormat.format(now)
			+ " " + name_
			+ " " + tag + ": " + msg);
	    ps_.flush();
	}
    }
    
    
}
