/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.util;

/**
 * A <code>Provider&lt;E&gt;</code> can provide a <code>Sink&lt;E&gt;</code>
 * with objects it reads from a <code>Source&lt;E&gt;</code>.
 * 
 * The class <code>Provider</code> implements a thread, which keeps reading from
 * the <code>Source</code> the <code>Provider</code> is associated with.
 * Whenever an object is read from this <code>Source</code>, this object is
 * written to the <code>Sink</code> the <code>Provider</code> is also associated
 * with.
 * 
 * <p>
 * 
 * Notice that once an instance of class <code>Provider</code> has been created,
 * the underlying thread must still be started explicitly by calling method
 * <code>start()</code> on this bbject.
 * 
 * <p>
 * 
 * Warning: The graceful termination of a <code>Provider</code> has not been
 * addressed yet.
 * 
 * @author $Author: F. Guidec$
 * @version $Revision: 1.0$
 */

public class Provider<E> 
extends Suspendable implements Runnable {

    protected Sink<E> sink_;

    protected Source<E> source_;
    
    public Verbosity verbosity = new Verbosity();
    
    /**
     * Create a new <code>Provider</code>, and associates this
     * provider with the specified <code>Source</code> and
     * <code>Sink</code>
     * 
     * @param source the <code>Source</code> this provider should
     *            read objects from
     * @param sink the <code>Sink</code> any object read from the
     *            source should be written to
     */
    public Provider(Source<E> source, Sink<E> sink) {
	setSource(source);
	setSink(sink);
    }
    
    /**
     * Returns a reference to the <code>Sink</code> this provider is associated
     * with
     * 
     * @return the <code>Sink</code> this provider is associated with
     */
    public Sink<E> getSink() {
	return sink_;
    }
    
    /**
     * Returns a reference to the <code>Source</code> this provider is
     * associated with
     * 
     * @return the <code>Source</code> this provider is associated with
     */
    public Source<E> getSource() {
	return source_;
    }
    
    /**
     * Reads once from the source and writes the result to the sink
     * 
     */
    public void proceed() throws Exception {
	
	if (verbosity.isEnabled()) {
	    System.out.println("Provider.proceed()");
	}
	
	E value;
	try {
	    value = source_.read();
	} catch (Exception e) {
	    if (verbosity.isEnabled())
		System.out.println("  Could not read from source " + source_
				   + e);
	    throw e;
	}
	
	if (verbosity.isEnabled())
	    System.out.println("  Read: " + value);
	
	if (isSuspended()) {
	    if (verbosity.isEnabled())
		System.out
		    .println("  Provider is suspended: discarding object read");
	    return;
	}
	
	if (value == null) {
	    if (verbosity.isEnabled())
		System.out
		    .println("  Aborting: the source returned a null object");
	    return;
	}
	
	try {
	    sink_.write(value);
	} catch (Exception e) {
	    if (verbosity.isEnabled())
		System.out.println("  Could not write to sink " + sink_);
	    throw e;
	}
    }
    
    /**
     * Starts this <code>Provider</code>. Once started, the
     * provider will keep reading objects from its source, and
     * writing the objects thus read to its sink.
     */
    public void run() {
	
	if (verbosity.isEnabled()) {
	    System.out.println("Provider.run()");
	}
	
	while (true) {
	    
	    try {
		proceed();
	    } catch (Exception e) {
		// e.printStackTrace();
	    }
	}
    }
    
    /**
     * Changes the <code>Sink</code> this provider should write to
     * 
     * @param sink
     *            the <code>Sink</code> this provider should write to
     */
    public void setSink(Sink<E> sink) {
	sink_ = sink;
    }
    
    /**
     * Changes the <code>Source</code> this provider should read from
     * 
     * @param source
     *            the <code>Source</code> this provider should read from
     */
    public void setSource(Source<E> source) {
	source_ = source;
    }
    
    /**
     * Returns a <code>String</code> representation of this
     * <code>Provider</code>
     * 
     * @return a <code>String</code> representation of this
     *         <code>Provider</code>
     */
    @Override
	public String toString() {
	return "Provider(source= " + source_ + ", sink=" + sink_ + ")";
    }
}
