/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/




package casa.dodwan.util;

import java.io.File;
import java.util.Hashtable;


/**
 * Provides access to properties attached to a service. The properties are Java
 * properties initialized from configuration files dedicated to this
 * service. For this purpose, a service is given a name that serves as key for
 * retrieving the configuration files. For the service named <code><em>service</em></code> on 
 * the DoDWAN host <code><em>host</em></code>, the property definitions are considered in 
 * the following order, a definition possibly being overriden by a subsequent one:
 * <p>
 * <ol>
 *  <li>Definitions in the default configuration file 
 *      (<code><em>conf</em>/<em>service</em>.defaults</code> in the home directory).</li>
 *  <li>Definitions in the user configuration file 
 *      (<code><em>host</em>/<em>conf</em>/<em>service</em>.conf</code> in the base directory)</li>
 *  <li>Definitions via the command line invocation of the JVM</li>
 * </ol>
 * See {@link SystemEnvironment} for details on the location of the home and base directories.
 */
public class ServiceEnvironment {
	
	// ------------------------------------------------------------
	private static Hashtable<String, ServiceEnvironment> instances = new Hashtable<String, ServiceEnvironment>();
	
	/**
     * @undocumented
     *  
	 * Give an instance of a subclass of ServiceEnvironment, identified by its service name.
	 * Subclasses may implement the singleton pattern.
	 * If a subclass doesn't implement the singleton pattern, give the last instance. 
	 */
    public static ServiceEnvironment getInstance(String serviceName) {
		return instances.get(serviceName);
	}

 // ------------------------------------------------------------
    private String name_ = null;
 
    /**
     *  @undocumented
     */
    public ConfigurationProperties properties = null;

    // ------------------------------------------------------------
    public ServiceEnvironment(String serviceName) {
	{
	    name_ = serviceName;
	    instances.put(serviceName, this);

	    
	    
	    String defaultConfFileName = System.getProperty(serviceName + ".defaults");
	    if (defaultConfFileName == null)  
		defaultConfFileName = SystemEnvironment.homePath  
		    + File.separator + "conf" + File.separator 
		    + serviceName
		    + ".defaults";
	    
	    String userConfFileName = System.getProperty(serviceName + ".conf");
	    if (userConfFileName == null)  {
	    	String userConfFileDirName = SystemEnvironment.baseHostPath 
	    			+ File.separator + "conf";
	    	File userConfFileDir = new File(userConfFileDirName);
	    	if (! userConfFileDir.exists()) userConfFileDir.mkdirs();
	    	userConfFileName = userConfFileDirName 
	    			+ File.separator
	    			+ serviceName 
	    			+ ".conf";
	    }

	    properties = new ConfigurationProperties(serviceName,
						     defaultConfFileName,
						     userConfFileName);
	}
    }

    // ------------------------------------------------------------
    /**
     * Returns the name of the service.
     * @return the name of the service.
     */
    public String getName() {
	return name_;
    }

    // ------------------------------------------------------------
    /**
     *  Returns the list of the attached configuration properties.
     * @return  the list of the attached configuration properties.
    */
    public ConfigurationProperties getProperties() {
	return properties;
    }
    
}
