/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.util;


/**
 * Defines operations on a component that can be enabled and disabled and that
 * accounts for external events of suspend/resume.  When the component is
 * enabled, it is fully operational. In the disabled state, it ceases to
 * fucntion and does not normally respond to invocations.  The component may be
 * informed of external events when the system is about to be suspended or when
 * the system resumes from a suspended state (typically when the host performs
 * a suspend to RAM).
 */
public class Suspendable {

	/**
	 * True if in enabled state, false otherwhise. Initially false. 
	 */
	protected boolean isEnabled_ = false;

	/**
	 * True if the last call to <code>suspend</code>/<code>resume</code> is
	 * a call to <code>resume</code>. Initially true.
	 */
	protected boolean isSuspended_ = true;

	/**
	 * Enters the disabled state.
	 */
	public void disable() {

		if (!isEnabled_)
			return;

		if (!isSuspended_)
			suspend();

		isEnabled_ = false;
	}

	/**
	 * Enters the enabled state.
	 */
	public void enable() {

		if (isEnabled_)
			return;

		isEnabled_ = true;
		//resume();
	}

	/**
	 * Checks if is enabled.
	 * 
	 * @return true, if is enabled
	 */
	public boolean isEnabled() {

		return isEnabled_;
	}

	/**
	 * @undocumented
	 *
	 * Checks if is suspended.
	 * 
	 * @return true, if is suspended
	 */
	public boolean isSuspended() {

		return isSuspended_;
	}

	/**
	 * Resume.
	 */
	public void resume() {

		if (isEnabled_)
			isSuspended_ = false;
	}

	/**
	 * Suspend.
	 */
	public void suspend() {

		if (isEnabled_)
			isSuspended_ = true;
	}

}
